//=============================================================================
// SoR_Outdoor_Weather.js
// MIT License (C) 2020 蒼竜 @soryu_rpmaker
// http://opensource.org/licenses/mit-license.php
// ----------------------------------------------------------------------------
// Latest version v1.03 (2020/08/11)
//=============================================================================


/*:ja
 * @plugindesc <室内天气失效> 禁用天气对室内处理的地图的影响。 
 * @author 蒼竜　@soryu_rpmaker
 *
 * 
 * @help
 * 事件命令“天气设置”可以让你在地图上创建雨雪效果，
 * 但是因为RPG Maker创建的地图本身没有“室内”或“室外”的概念，
 * 所以是“室内”。出现“即使我在室内也在下雨”等现象。 
 * 
 * 即使根据地图的进入和退出执行“设置天气”命令，
 * 如果在某个开关打开时根据标志信息更改天气，
 * 则编辑器的管理也会变得复杂。
 * 
 * 该插件通过预先指定要在室外处理的地图的 ID，
 * 在您在室内地图上时禁用天气设置。
 * （当您外出时，将再次恢复设定的天气。） 
 *
 * -----------------------------------------------------------
 * 用法（插件设置） 
 * -----------------------------------------------------------
 * 
 *　在插件管理中，指定SoR_Outdoor_Weather参数Map_IDs的ID为“outdoor”
 * （希望受天气设置影响的地图）。 
 *  它支持使用连字符的逗号分隔和序列号输入。 
 *
 * 示例 1) 地图编号 30 
 * 30
 * 示例 2) 4到8号地图
 * 4-8
 * 示例 3) 地图 2、5、6 和 10 到 13 和地图 100 
 * 2,5,6,10-13,100
 *
 *
 * 请照常调用事件命令“天气设置”。
 * 室外地图采用RPG工具的默认行为，
 * 只有在室内移动时才能禁用天气的影响。
 *
 * 如果您在室内地图上调用“天气设置”，
 * 这些设置将一直储存到您外出为止，并会在您外出时反映出来。 
 *
 *
 * 详细例子 
 * http://dragonflare.dip.jp/dcave/
 * 它在中描述。 
 *
 *
 * -----------------------------------------------------------
 * バージョン情報
 * -----------------------------------------------------------
 * v1.03 (2020/08/11)       SoR命名規則統一(機能変更はありません)
 * v1.02 (2020/06/04)      プラグインコマンド処理修正  
 * v1.01 (2020/05/20)      天候の戦闘シーンへの反映オプション追加  
 * v1.00 (2020/05/14)       公開  
 *
 * @param Map_IDs
 * @desc 将被视为户外的地图ID（因“天气设置”而发生变化的地图 ID） 
 * @default 1-3
 *
 * @param IsBattleWeather_enable
 * @desc 当'true'时，天气设置将反映在战斗场景中。 
 * @default true
 * @type boolean
 */
 /*:
 * @plugindesc <禁用室内地图的天气效果> 将指定地图中的天气效果命令禁用为室内。 
 * @author Soryu @soryu_rpmaker
 *
 *
 * @help 
 * ------------------------------------------------------------
 * 如何使用 
 * ------------------------------------------------------------
 *　在插件管理中，指定SoR_Outdoor_Weather参数Map_IDs的ID为“outdoor”
 * （希望受天气设置影响的地图）。 
 *  它支持使用连字符的逗号分隔和序列号输入。 
 *
 * 例如）对于 ID 为 30 的地图 
 * 30
 *
 * 例如）对于 ID 为 4 到 8 的地图 
 * 4-8
 *
 * 例如）对于 ID 为 2、5、6、10 到 13 和 100 的地图 
 * 2,5,6,10-13,100
 *
 * 在 RPGMV 中，您可以像往常一样调用天气设置的事件命令。
 * 如果玩家在户外，您的游戏将采用 RPGMV 中的默认行为。
 * 在室内时，天气设置被禁用。
 * 移至室外后，天气恢复。 
 * 
 * 请注意，如果您在室内调用天气设置的事件命令，
 * 天气设置将在玩家进入室外后（临时存储和）应用。 
 * 
 *
 * 要获取更多说明，请参阅  http://dragonflare.dip.jp/dcave/ .
 *
 *
 * ------------------------------------------------------------
 * Version info
 * ------------------------------------------------------------
 * v1.03 (Aug 11, 2020)      Unified the regulation of naming SoR variables (no functions are modified)
 * v1.02 (Jun 04, 2020)       Modified for a plugin command
 * v1.01 (May 20, 2020)       An option enables weather effect in battle scene is implemented.
 * v1.00 (May 14, 2020)       Released!
 *
 * @param Map_IDs
 * @desc 将被视为户外的地图ID（因“天气设置”而发生变化的地图 ID）
 * @default 1-3
 *
 * @param IsBattleWeather_enable
 * @desc 如果为“true”，天气设置也适用于战斗场景。 
 * @default true
 * @type boolean
 */
 



(function () {

  var Parameters = PluginManager.parameters('SoR_Outdoor_Weather');
  var IsBattleWeather_enable = Boolean(Parameters['IsBattleWeather_enable'] === 'true' || false);
  
	var SoR_OW_GT_initialize = Game_Temp.prototype.initialize;
	Game_Temp.prototype.initialize = function() {
		this._weather_disableIDs = convParam2MapID(Parameters);
		this._weather_disable_flag = false;
		this._weather_enable_inBattle = IsBattleWeatherEnable(Parameters);
		this._weather_stack = [];
		
		SoR_OW_GT_initialize.call(this);
	};

	var SoR_OW_GM_setup = Game_Map.prototype.setup;
	Game_Map.prototype.setup = function(mapId){
		SoR_OW_GM_setup.call(this, mapId);

		var i;
		var len = $gameTemp._weather_disableIDs.length;
		for(i=0;i<len;i++){
		   if( $gameMap.mapId() == $gameTemp._weather_disableIDs[i] ){
			 $gameTemp._weather_disable_flag = false;
			 break;
		   }
		}
		if(i == len) $gameTemp._weather_disable_flag = true;
		
		//move to outdoor && with weather change 
		// (-> restore weather settings)
		if(!$gameTemp._weather_disable_flag && $gameTemp._weather_stack.length != 0 ){
			var type = $gameTemp._weather_stack.shift();
			var power = $gameTemp._weather_stack.shift();
			$gameScreen.changeWeather(type, power, 1);
		}
		//move to indoor && with weather change 
		// (-> store weather settings and disable current weather settings)	
		if($gameTemp._weather_disable_flag && $gameScreen._weatherType !='none'){
			$gameTemp._weather_stack.push($gameScreen._weatherType);
			$gameTemp._weather_stack.push($gameScreen._weatherPower);		
			$gameScreen.changeWeather('none',0,0,1);
		}
	}
	
	function IsBattleWeatherEnable(param){
	  if (IsBattleWeather_enable) return true;
      else return false;
	}
	
	function convParam2MapID(param){
		var str = param['Map_IDs'];
		var id_array = [];
		var spl_str = str.split(',');
		var regex = /(\d+)\s*-\s*(\d+)/;

		if (!str) return id_array; //empty
		
		for (var i = 0; i < spl_str.length; i++){
		   if (regex.test(spl_str[i]) == true){
			 var begin = Number(RegExp.$1);
			 var end = Number(RegExp.$2);
			 for (var j = begin; j <= end; j++) id_array.push(j); //ranged ID
		   }
		   else id_array.push(Number(spl_str[i])); //one ID
		}
		
	  return id_array;
	}


//Duplicate process in Spriteset_Map class
  var SoR_OW_SB_createLowerLayer = Spriteset_Battle.prototype.createLowerLayer;
  Spriteset_Battle.prototype.createLowerLayer = function() {
    SoR_OW_SB_createLowerLayer.call(this);
    if($gameTemp._weather_enable_inBattle) this.createWeather();
  }

  Spriteset_Battle.prototype.createWeather = function() {
    this._weather = new Weather();
    this.addChild(this._weather);
  }
  
  var SoR_OW_SB_update = Spriteset_Battle.prototype.update;
  Spriteset_Battle.prototype.update = function() {
    SoR_OW_SB_update.call(this);
    if($gameTemp._weather_enable_inBattle) this.updateWeather();
  }
  
  Spriteset_Battle.prototype.updateWeather = function() {
    this._weather.type = $gameScreen.weatherType();
    this._weather.power = $gameScreen.weatherPower();
    this._weather.origin.x = 0;
    this._weather.origin.y = 0;
  }


// Set Weather Effect
Game_Interpreter.prototype.command236 = function() {
    if ( $gameTemp._weather_disable_flag == false){
		$gameScreen.changeWeather(this._params[0], this._params[1], this._params[2]);
        if(this._params[3] == true) this.wait(this._params[2]); //wait flag
    }
	else if($gameTemp._weather_disable_flag == true){ // reserve weather change until go outdoor
		$gameTemp._weather_stack.length = 0;
		$gameTemp._weather_stack.push(this._params[0]);
		$gameTemp._weather_stack.push(this._params[1]);	
	}
    return true;
}

})();