//=============================================================================
// TAA_EnemyDrops.js
// Author: taaspider
//=============================================================================

var TAA = TAA || {};
TAA.ed = {};
TAA.ed.Version = "1.0.1";
TAA.ed.PluginName = "TAA_EnemyDrops";
TAA.ed.alias = {};
TAA.ed.trace = 0;

/*:
 * @target MV MZ
 * 
 * @plugindesc [1.0.1] Enemy drops advanced customization
 * @author T. A. A. (taaspider)
 * @url http://taaspider.itch.io/ 
 * 
 * @help
 * ============================================================================
 * Terms of Use
 * ============================================================================
 * Any plugins developed by taaspider are free for use for both commercial and 
 * noncommercial RPG Maker games, unless specified otherwise. Just remember to
 * credit "Taaspider".
 * 
 * Redistribution of parts or the whole of taaspider plugins is forbidden, unless
 * it comes from the official website: http://taaspider.itch.io. You are allowed 
 * to edit and change the plugin code for your own use, but you're definitely not 
 * allowed to sell or reuse any part of the code as your own. Although not 
 * required to use my plugins, a free copy of your game would be nice!
 * 
 * If you enjoy my work, consider offering a donation when downloading my plugins, 
 * or offering a monthly pledge to my Patreon account. It would be of great help!
 * Also, follow me on facebook to get firsthand news on my activities!
 *  Facebook: https://www.facebook.com/taaspider 
 *  Patreon: https://www.patreon.com/taaspider
 * 
 * =============================================================================
 * Introduction
 * =============================================================================
 * 
 * WARNING: This plugin requires RPG Maker MV 1.5.0 or above! Please make sure 
 * your RPG Maker MV software is up to date before using this plugin.
 * You don't need any specific version if you're using MZ.
 * 
 * -----------------------------------------------------------------------------
 * 
 * This plugin aims to give the developer more control and flexibility when 
 * setting up loot drops from enemies. With it you can set advanced rules to 
 * influence drop probability, determine the number of times a specific item can 
 * be dropped and how probability increases/decreases after each drop, which means 
 * a second drop can be more or less likely to happen. You can also influence 
 * probability with a variable, so that drop rate can be affected without much work.
 * 
 * As of most of my plugins until now, EnemyDrops can be configured either through 
 * an external JSON file, plugin parameters or enemy note tags. It's up to the dev 
 * to select the method he's most comfortable with! Also, for those who use an 
 * external JSON file, you have the option to use a single file for both 
 * TAA_EnemyLevels and TAA_EnemyDrops, centralizing enemy setup in a single place 
 * for both plugins (and maybe others in the future).
 * 
 * There's a huge list of possible comparisons and conditionals to compose
 * a drop rule, be sure to read the help file thoroughly to check on all 
 * possibilities!
 * 
 * =============================================================================
 * DataSource Config
 * =============================================================================
 * 
 * Enemy drop conditional rules can be defined for each enemy using three different 
 * methods: an external JSON file, using Plugin Parameters, and through enemy note 
 * tags. You can't use both a JSON file and Plugin Parameters at the same time, so 
 * you need to choose one through the parameter "SourceType". 
 * 
 * Note tags an be used regardless of which one of the two you choose, and has a 
 * higher priority than both. That means if an enemy have any drop rules set in a
 * JSON file, for example, but also have note tags, the note tags will overwrite
 * whatever is loaded from the file (regardless of it being applied to the same
 * item or not).
 * 
 * You can define if engine drop settings are still considered by setting the 
 * Plugin parameter 'Engine Settings' to either 'Keep' (keep both whatever is 
 * configured in the engine and conditional drops) or 'Discard' (discard anything
 * set in the engine and use only conditional drops). This option can be useful
 * so you can keep using the engine for simple drop rules (like an enemy that
 * always drop a Potion, for example).
 * 
 * OBS.: The JSON file can be shared with other plugins from the TAA library 
 * (typically TAA_Enemy* plugins, like TAA_EnemyLevels), so you can concentrate 
 * all enemy config data in one place. Be sure to check other plugins help 
 * sections and see which one supports this datasource sharing.
 * 
 * -----------------------------------------------------------------------------
 * JSON Config
 * -----------------------------------------------------------------------------
 * 
 * File
 *  - This should point to a JSON file containing your enemy conditional drop rules. 
 * The file must be placed inside your game's data folder.
 * The JSON file must have a structure similar to this:
 *  [
 *      {
 *          "id": <enemy ID>,
 *          "drops": [
 *              {
 *                  "item": <item id / name (without blank spaces)>,
 *                  "armor": <armor id / name (without blank spaces)>,
 *                  "weapon": <weapon id / name (without blank spaces)>,
 *                  "type": <drop type>,
 *                  "base": <base drop chance>,
 *                  "amount": <drop amount>
 *                  "iterationEffect": {
 *                      "type": <type code>,
 *                      "attenuation": <attenuation value>
 *                  },
 *                  "conditions": [
 *                      <conditional clause 1>,
 *                      <conditional clause 2>,
 *                      ...
 *                  ]
 *              },
 *              {
 *                  ...
 *              },
 *              ...
 *          ]
 *      }
 *  ]
 * 
 * This plugin supports sharing the datasource file with others from the TAA Library,
 * like TAA_EnemyLevels. To use this feature, simply configure all plugins (which support
 * this) to use the same JSON file. If you prefer keeping things separate, just config
 * TAA_EnemyDrops to use a different file altogether and the plugin will merge data on
 * game boot.
 * 
 * You can also mix datasource types with different plugins without worries (for example,
 * using a JSON File for TAA_EnemyLevels and Plugin Manager data for TAA_EnemyDrops).
 * 
 * You can customize the object's tags as you wish using the next parameters.
 * 
 * IMPORTANT: 
 *  - <enemy ID> must be a valid enemy ID in the database, and must be unique in
 * your JSON file. More than one entry for the same id can (and will) generate 
 * errors.
 *  - <item id> should be replaced either by the item id from the database, or its
 * name. If you choose to use an item name, be sure to surround it with quotes and
 * remove any blank spaces. For example: an item named Big Potion can be referenced
 * here as "BigPotion". You need at least one object from item, armor or weapon for
 * each rule, but it is not necessary to have all three present.
 *  - <armor id> should be replaced either by the armor id from the database, or
 * its name. The same instructions given above for item names are applied. You need 
 * at least one object from item, armor or weapon for each rule, but it is not 
 * necessary to have all three present.
 *  - <weapon id> should be replace either by the weapon id from the database, or
 * its name. The same instructions given above for item names are applied. You need 
 * at least one object from item, armor or weapon for each rule, but it is not 
 * necessary to have all three present.
 *  - <drop type> defines what the drop rules refer to. It has four possible values:
 *      "item": only the item object is effectively dropped (if conditions are met). If
 *              this type is set, the "item" object must be present.
 *      "armor": only the armor object is effectively dropped (if conditions are met). If
 *              this type is set, the "armor" object must be present.
 *      "weapon": only the weapon object is effectively dropped (if conditions are met). If
 *              this type is set, the "weapon" object must be present.
 *      "all": any item, armor or weapon object present is effectively dropped (if
 *             conditions are met). At least one of the drop objects must be present
 *             ("item", "armor", "weapon")
 *  - <base drop chance> is a value between 0 and 100 and represents how likely the
 * drop is to happen. It is a percentage value, and the higher it is, the more likely
 * it succeeds.
 *  - <drop amount> is the number of times the item / armor / weapon should be 
 * dropped when conditions are met. You can define how drop chance applies for
 * multiple item / armor / weapons instances with the iterationEffect object.
 *  - <type code> is a numeric value representing one of three possible behaviors:
 *         0: Shared probability. That means that the drop chance is tested a single
 *            time for all items. If it pass, all items are dropped. If it fails,
 *            none are.
 *         1: Repeat probability. That means that each item has its own chance of
 *            being dropped, but the same probability is applied to each one of them.
 *         2: Attenuate probability. That means that for each item instance evaluated,
 *            drop probability for the next one is reduced/increased by the <attenuation 
 *            value>. This applies individually to items, armors and weapons. So if
 *            a drop rule is set to drop 2 of each, the conditional rules calculate
 *            the initial probability for the first item, armor and weapon, but the
 *            second of each one of them is attenuated.
 *  - <attenuation value> is a number between -100 and 100 and represents how much to
 * attenuate (or increase the chance) of a second drop when the drop amount is greater
 * than 1 and <type code> is set to 2.
 *  - <conditional clause> are evaluation clauses, as described in the section "Conditional 
 * Clauses".
 * 
 * -----------------------------------------------------------------------------
 * Plugin Manager Data
 * -----------------------------------------------------------------------------
 * 
 * Setting up enemy level data through Plugin Parameters is pretty straightforward,
 * just make sure there is only one entry for each enemy ID.
 * 
 * Also keep in mind that if the plugin TAA_EnemyLevels is present and enabled, 
 * TAA_EnemyDrops is forced to use the same datasource type. If EnemyLevels is set
 * to use an external JSON file, EnemyDrops will need to use the same file with
 * all settings bundled together. If it is set to use the Plugin Manager, EnemyDrops
 * will need to be set through the Plugin Manager as well.
 * 
 * Enemy note tags can be used regardless of both plugin configs!
 * 
 * -----------------------------------------------------------------------------
 * Conditional Clauses
 * -----------------------------------------------------------------------------
 * 
 * This section covers all possible conditional clauses that can increase or decrease
 * drop probability for each drop rule. All drop rules follows the same structure:
 *      <condition>:<result>:<stop clause>
 *  <result> represents how the drop probability is affected if the conditional clause
 *      is met, and it can assume one of two formats:
 *      <op><value>%
 *          <op> is either + (to increase the probability), - (to decrease it), * (to
 *          multiply it) or = to set it to a specific value.
 *          <value> is a percentage, and must be a value between 0 and 100 (decimals 
 *          can be used for higher precision after the dot. Example: 0.5%)
 *      <op><value>
 *          Here <op> has the same possible values, but <value> is treated as a floating
 *          representation of a percentage value. It must be a value between 0 and 1, 
 *          where 0 represents 0%, and 1 represents 100%. To represent a 50% value, for
 *          example, you would write 0.5 without the '%' character. Be careful, as writing
 *          0.5% is way different than typing 0.5.
 *  <stop clause> is an optional clause which can determine if the conditional evaluation
 *      should stop depending on the conditional clause result. It must be wither one of
 *      two options:
 *          'stop if false': If the condition is not met, do not evaluate remaining clauses
 *          'stop if true': If the condition is met, do not evaluate remaining clauses
 *      This option can be useful if you require one condition to be met to evaluate
 *      other conditions. For example, if the actor's level is less than 10, don't bother
 *      testing other conditions.
 *  <condition> is the conditional clause itself. All possible clauses are described
 *      bellow.
 * 
 * VALUE COMPARISON
 * ----------------
 * <value> <op> <value>
 *  <value> can be either a pure integer, or a variable value. To access a variable,
 *  type in v[n], where n is the variable number.
 *  <op> is a comparison operator. Here's a list of what's supported:
 *      >:  greater than
 *      >=: greater or equal than
 *      <: less than
 *      <=: less or equal than
 *      ===: equal to
 *      !==: different than
 * 
 * Examples:
 *  v[1] >= 10
 *  v[2] === v[3]
 * 
 * SWITCH STATE
 * ------------
 * <switch> <statement> <state>
 * <switch> <op> <switch>
 *  <switch represents a switch value, and must be type as s[n], where n is the
 *  switch number.
 *  <statement> is a text to evaluate the current state of the switch. It must be
 *  either 'is' or 'is not'. If you use 'is', the conditional will succeed if the
 *  switch has the specified value. If using 'is not', it will succeed if the switch
 *  has the opposite value.
 *  <state> is either true or false, and represents the possible states of a switch.
 *  <op> is a limited version of the value comparison. Here only two possible operators
 *  are allowed:
 *      ===: equal to
 *      !==: different than
 * 
 * Examples:
 *  s[1] is true
 *  s[2] is not false
 *  s[4] === s[3]
 * 
 * ACTOR CLAUSES
 * -------------
 * ACTOR <actor> <statement> <state>
 *  This clause can be used to check if a specific actor is either alive or dead.
 *  <actor> is either the actor ID, a variable referencing the actor ID (using the 
 *  v[n] format) or the actor name itself. If the actor name has blank spaces in it,
 *  here it should be typed without it. For example, if your character name is 'King
 *  Richard The III', you can reference him as 'KingRichardTheIII'.
 *  <statement> is the same as with the switch state clause. It is either 'is' or
 *  'is not'.
 *  <state> is either ALIVE or DEAD
 * 
 * Examples:
 *  ACTOR 1 IS ALIVE
 *  ACTOR Therese IS NOT DEAD
 * 
 * ACTOR <actor> <not> IN PARTY
 *  This clause can be used to check if a specific actor is part of the current 
 *  party. The <not> clause can be used to negate the evaluation and is completely
 *  optional. It being present or not defines if we're checking if the actor is in
 *  or out of the party.
 * 
 * Examples:
 *  ACTOR Lucius IN PARTY
 *  ACTOR 2 NOT IN PARTY
 * 
 * ACTOR <actor> <not> STATE <state>
 *  This clause can be used to check if a specific actor is or is not affected by
 *  a particular state.
 *  <state> can be either a state ID, a variable referencing the state ID (using the
 *  v[n] format) or the state name itself. If the state name has blank spaces in it,
 *  here it should be typed without it. For example, 'Def Up' should be entered as
 *  'DefUp'.
 * 
 * Example:
 *  ACTOR Harold NOT STATE Blind
 *  ACTOR v[1] STATE DefUp
 *  ACTOR v[2] NOT STATE v[1]
 * 
 * ACTOR <actor> <not> CLASS <class>
 *  This clause can be used to check if a specific actor has or has not a particular 
 *  class.
 *  <class> can be either a class ID, a variable referencing the class ID (using
 *  the v[n] format) or the class name itself (with no blank spaces).
 * 
 * Examples:
 *  ACTOR 1 CLASS Hero
 *  ACTOR v[1] NOT CLASS GreatMage
 *  ACTOR Arnold CLASS v[5]
 * 
 * ACTOR <actor> <not> ARMOR <armor>
 *  This clause can be used to check if a specific actor has or has not a particular
 *  armor equipped.
 *  <armor> can be either an armor ID, a variable referencing the armor ID (using the
 *  v[n] format) or the armor name itself (with no blank spaces).
 * 
 * Examples:
 *  ACTOR 1 ARMOR LeatherArmor
 *  ACTOR v[2] NOT ARMOR 1
 *  ACTOR Harold NOT ARMOR v[4]
 * 
 * ACTOR <actor> <not> ARMOR TYPE <type>
 *  This clause can be used to check if a specific actor has or has not a particular
 *  armor type equipped.
 *  <type> can be either an armor type ID, a variable referencing the armor type ID
 *  (using the v[n] format) or the armor type name itself (with no blank spaces).
 * 
 * Examples:
 *  ACTOR 1 ARMOR TYPE HeavyArmor
 *  ACTOR v[34] NOT ARMOR TYPE v[1]
 *  ACTOR Lucius ARMOR TYPE 3
 * 
 * ACTOR <actor> <not> WEAPON <weapon>
 *  This clause can be used to check if a specific actor has or has not a particular
 *  weapon equipped.
 *  <weapon> can be either a weapon ID, a variable referencing the weapon ID (using the
 *  v[n] format) or the weapon name itself (with no blank spaces).
 * 
 * Examples:
 *  ACTOR 1 NOT WEAPON 5
 *  ACTOR Therese WEAPON BlackSword
 *  ACTOR v[1] NOT WEAPON v[5]
 * 
 * ACTOR <actor> <not> WEAPON TYPE <type>
 *  This clause can be used to check if a specific actor has or has not a particular
 *  weapon type equipped.
 *  <type> can be either an weapon type ID, a variable referencing the weapon type ID
 *  (using the v[n] format) or the weapon type name itself (with no blank spaces).
 * 
 * Examples:
 *  ACTOR 1 WEAPON TYPE Dagger
 *  ACTOR v[34] NOT WEAPON TYPE v[1]
 *  ACTOR Lucius WEAPON TYPE 3
 * 
 * ACTOR <actor> <not> SKILL <skill>
 *  This clause can be used to check if a specific actor has or has not a particular
 *  skill available.
 *  <skill> can be either a skill ID, a variable referencing the skill ID (using the
 *  v[n] format) or the skill name itself (with no blank spaces).
 * 
 * Examples:
 *  ACTOR 1 SKILL 4
 *  ACTOR Harold NOT SKILL LuckyStrike
 *  ACTOR v[4] SKILL v[5]
 * 
 * ACTOR <actor> <not> SKILL <skill> LEVEL <level>
 *  This clause can be used to check if a specific actor has or has not a particular
 *  skill available at an specified level. This clause requires TAA_SkillMastery to
 *  work. If it is used without the required plugin the level is ignored and conditions
 *  will be met depending on the actor having the skill or not.
 *  <level> can be either a number or a variable value (using the v[n] format).
 * 
 * Examples:
 *  ACTOR 1 SKILL 4 LEVEL 2
 *  ACTOR Harold SKILL DoubleStrike LEVEL v[1]
 *  ACTOR v[2] SKILL v[2] LEVEL 4 
 * 
 * ACTOR <actor> <not> LEVEL <op> <level>
 *  This clause can be used to act according to a specific actor level.
 *  <op> represents all conditional operators listed for VALUE COMPARISON.
 * 
 * Examples:
 *  ACTOR Harold NOT LEVEL < 5
 *  ACTOR Harold LEVEL >= 5
 *  ACTOR v[2] LEVEL !== 3
 * 
 * PARTY CLAUSES
 * -------------
 * 
 * PARTY <not> STATE <state>
 *  This clause can be used to check if any member of the current party is affected
 *  by a given state.
 *  <state> can be either a state ID, a variable referencing the state ID (using the
 *  v[n] format) or the state name itself. If the state name has blank spaces in it,
 *  here it should be typed without it. For example, 'Def Up' should be entered as
 *  'DefUp'.
 * 
 * Examples:
 *  PARTY STATE LuckUp
 *  PARTY NOT STATE v[2]
 *  PARTY STATE 8
 * 
 * PARTY <not> CLASS <class>
 *  This clause can be used to check if any member of the current party is part of
 *  a given class.
 *  <class> can be either a class ID, a variable referencing the class ID (using
 *  the v[n] format) or the class name itself (with no blank spaces).
 * 
 * Examples:
 *  PARTY CLASS 2
 *  PARTY NOT CLASS Hero
 *  PARTY CLASS v[5]
 * 
 * PARTY <not> ARMOR <armor>
 *  This clause can be used to check if any member of the current party have a
 *  particular armor equipped.
 *  <armor> can be either an armor ID, a variable referencing the armor ID (using the
 *  v[n] format) or the armor name itself (with no blank spaces).
 * 
 * Examples:
 *  PARTY ARMOR Helmet
 *  PARTY NOT ARMOR 3
 *  PARTY ARMOR v[5]
 * 
 * PARTY <not> ARMOR TYPE <type>
 *  This clause can be used to check if any member of the current party have a
 *  particular armor type equipped.
 *  <type> can be either an armor type ID, a variable referencing the armor type ID
 *  (using the v[n] format) or the armor type name itself (with no blank spaces).
 * 
 * Examples:
 *  PARTY NOT ARMOR TYPE 2
 *  PARTY ARMOR TYPE v[2]
 *  PARTY ARMOR TYPE LuckyGarment
 *  
 * PARTY <not> WEAPON <weapon>
 *  This clause can be used to check if any member of the current party have a 
 *  particular weapon equipped.
 *  <weapon> can be either a weapon ID, a variable referencing the weapon ID (using the
 *  v[n] format) or the weapon name itself (with no blank spaces).
 * 
 * Examples:
 *  PARTY NOT WEAPON RustSword
 *  PARTY WEAPON 5
 *  PARTY WEAPON v[4]
 * 
 * PARTY <not> WEAPON TYPE <type>
 *  This clause can be used to check if any member of the current party have a
 *  particular weapon type equipped.
 *  <type> can be either an weapon type ID, a variable referencing the weapon type ID
 *  (using the v[n] format) or the weapon type name itself (with no blank spaces).
 * 
 * Examples:
 *  PARTY NOT WEAPON TYPE 5
 *  PARTY WEAPON TYPE v[4]
 *  PARTY WEAPON TYPE Dagger
 * 
 * PARTY <not> SKILL <skill>
 *  This clause can be used to check if any member of the current party have a
 *  particular skill available.
 *  <skill> can be either a skill ID, a variable referencing the skill ID (using the
 *  v[n] format) or the skill name itself (with no blank spaces).
 * 
 * Examples:
 *  PARTY SKILL 4
 *  PARTY NOT SKILL LuckyStrike
 *  PARTY SKILL v[5]
 * 
 * PARTY <not> SKILL <skill> LEVEL <level>
 *  This clause can be used to check if any member of the current party have a particular
 *  skill available at an specified level. This clause requires TAA_SkillMastery to
 *  work. If it is used without the required plugin the level is ignored and conditions
 *  will be met depending on the actor having the skill or not.
 *  <level> can be either a number or a variable value (using the v[n] format).
 * 
 * Examples:
 *  PARTY SKILL 4 LEVEL 2
 *  PARTY SKILL DoubleStrike LEVEL v[1]
 *  PARTY SKILL v[2] LEVEL 4 
 * 
 * PARTY <not> LEVEL <type> <op> <level>
 *  This clause can be used to act according to party members levels.
 *  <type> can assume three possible values:
 *      AVG: Comparison will take place with the average battle members levels
 *      MAX: Comparison will take place with the highest battle member level
 *      MIN: Comparison will take place with the lowest battle member level
 *  Alternatively, <type> can be completely omitted, which will cause the plugin to
 *  evaluate the comparison against all battle members. If any member meets the
 *  requirements the condition is met.
 *  <op> represents all conditional operators listed for VALUE COMPARISON.
 *  <level> can be either a number or a variable value (using the v[n] format).
 * 
 * Examples:
 *  PARTY NOT LEVEL AVG < 5
 *  PARTY LEVEL >= v[2]
 *  PARTY MAX LEVEL <= 10
 * 
 * PARTY <not> <type> MEMBERS <op> <value>
 *  This clause can be used to act according to the party size in multiple occasions,
 *  depending on <type> value.
 *  <type> can assume one of three values:
 *      ALIVE: Comparison will take place with the number of party members that are
 *          still alive;
 *      DEAD: Comparison will take place with the number of party members that are
 *          currently dead;
 *      BATTLE: Comparison will take place considering how many battle members we
 *          have in the party;
 *  Alternatively, <type> can be completely omitted, which will cause the plugin to
 *  evaluate the comparison considering the whole party size (including non-battle
 *  members).
 *  <op> represents all conditional operators listed for VALUE COMPARISON.
 *  <value> can be either a number or a variable value (using the v[n] format).
 * 
 * Examples:
 *  PARTY NOT MEMBERS > 4
 *  PARTY MEMBERS <= v[4]
 *  PARTY DEAD MEMBERS < 2
 *  PARTY NOT ALIVE MEMBERS < v[3]
 *  PARTY BATTLE MEMBERS > 3
 * 
 * PARTY <not> GOLD <op> <value>
 *  This clause can be used to check the gold currently held by the party.
 *  <op> represents all conditional operators listed for VALUE COMPARISON.
 *  <value> can be either a number or a variable value (using the v[n] format).
 * 
 * Examples:
 *  PARTY GOLD >= 1000
 *  PARTY NOT GOLD < v[5]
 * 
 * ENEMY CLAUSES
 * -------------
 * 
 * ENEMY <not> DEATH TURN <op> <value>
 *  This clause can be used to act according to which turn the enemy dies.
 *  If an enemy is defeated and brought back by any skill, this will compare
 *  to its last death (or to each one separately, depending on how drops are
 *  configured within the plugin).
 *  <op> represents all conditional operators listed for VALUE COMPARISON.
 *  <value> can be either a number or a variable value (using the v[n] format).
 * 
 * Examples:
 *  ENEMY DEATH TURN <= 5
 *  ENEMY NOT DEATH TURN > v[3]
 * 
 * ENEMY <not> DEATH DAMAGE <op> <value>
 *  This clause can be used to act according to how much damage was dealt at
 *  the killing blow.
 *  <op> represents all conditional operators listed for VALUE COMPARISON.
 *  <value> can be either a number or a variable value (using the v[n] format).
 * 
 * Examples:
 *  ENEMY NOT DEATH DAMAGE < 100
 *  ENEMY DEATH DAMAGE >= v[6]
 * 
 * ENEMY <not> LEVEL <op> <level>
 *  This clause can be used to act according to the enemy level. This requires
 *  TAA_EnemyLevels to work. If the plugin is not present the clause is ignored.
 *  <op> represents all conditional operators listed for VALUE COMPARISON.
 *  <level> can be either a number or a variable value (using the v[n] format).
 * 
 * Examples:
 *  ENEMY NOT LEVEL < 10
 *  ENEMY LEVEL >= v[5]
 * 
 * ENEMY <not> STRUCK LAST BY <type> <value>
 *  This clause can be used to act according to what has struck the enemy last
 *  before its death.
 *  <type> can assume one of the following options:
 *      SKILL: Last skill to hit the enemy;
 *      ACTOR: Last actor to hit the enemy;
 *      ELEMENT: Last element to hit the enemy;
 *      STYPE: Last skill type to hit the enemy;
 *      ITEM: Last item to hit the enemy;
 *  <value> is either an ID (skill, actor, element, skill type or item, according
 *  to <type>), a variable referencing an ID (using the v[n] format) or the <type>
 *  name itself (without blank spaces).
 * 
 * Examples:
 *  ENEMY STRUCK LAST BY SKILL Fireball
 *  ENEMY NOT STRUCK LAST BY ACTOR 1
 *  ENEMY STRUCK LAST BY ELEMENT v[5]
 * 
 * ENEMY <not> STRUCK COUNT BY <type> <type value> <op> <value>
 *  This clause can be used to act according to how many times <type> has hit the
 *  enemy during battle.
 *  <type> can assume one of the following options:
 *      SKILL: Last skill to hit the enemy;
 *      ACTOR: Last actor to hit the enemy;
 *      ELEMENT: Last element to hit the enemy;
 *      STYPE: Last skill type to hit the enemy;
 *      ITEM: Last item to hit the enemy;
 *  <type value> is either an ID (skill, actor, element, skill type or item, according
 *  to <type>), a variable referencing an ID (using the v[n] format) or the <type>
 *  name itself (without blank spaces).
 *  <op> represents all conditional operators listed for VALUE COMPARISON.
 *  <value> can be either a number or a variable value (using the v[n] format).
 * 
 * Examples:
 *  ENEMY NOT STRUCK COUNT BY SKILL 1 > 10
 *  ENEMY STRUCK COUNT BY STYPE Normal < 5
 *  ENEMY STRUCK COUNT BY ITEM v[6] > v[5]
 * 
 * ENEMY <not> SKILL LAST <skill>
 *  This clause can be used to check the last skill used by the enemy before death.
 *  <skill> can be either a skill ID, a variable referencing the skill ID (using the
 *  v[n] format) or the skill name itself (with no blank spaces).
 * 
 * Examples:
 *  ENEMY NOT SKILL LAST 5
 *  ENEMY SKILL LAST Drain
 *  ENEMY SKILL LAST v[7]
 *  
 * ENEMY <not> SKILL COUNT <skill> <op> <value>
 *  This clause can be used to check how many times the enemy has used a particular
 *  skill.
 *  <skill> can be either a skill ID, a variable referencing the skill ID (using the
 *  v[n] format) or the skill name itself (with no blank spaces).
 *  <op> represents all conditional operators listed for VALUE COMPARISON.
 *  <value> can be either a number or a variable value (using the v[n] format).
 * 
 * Examples:
 *  ENEMY NOT SKILL COUNT 1 > 10
 *  ENEMY SKILL COUNT Attack < v[3]
 *  ENEMY SKILL COUNT v[1] === v[9]
 * 
 * ITEM CLAUSES
 * -------------
 * 
 * ITEM <not> LAST <item>
 *  This clause can be used to check which item has been used last by the party.
 *  <item> can be either a item ID, a variable referencing the item ID (using the
 *  v[n] format) or the item name itself (with no blank spaces).
 * 
 * Examples:
 *  ITEM NOT LAST 5
 *  ITEM LAST LuckyPowder
 * 
 * ITEM <not> TIMES USED <item> <op> <value>
 *  This clause can be used to act according to how many times the party have used
 *  a particular item.
 * 
 * Examples:
 *  ITEM TIMES USED Potion < 2
 *  ITEM NOT TIMES USED v[4] > v[7]
 *  ITEM TIMES USED 1 <= 5
 * 
 * ITEM <not> OWNS <item> <op> <value>
 *  This clause checks how many of a certain item can be found in the party's 
 *  inventory.
 *  <op> represents all conditional operators listed for VALUE COMPARISON.
 *  <value> can be either a number or a variable value (using the v[n] format).
 * 
 * Examples:
 *  ITEM OWNS Potion < 100
 *  ITEM NOT OWNS v[4] > 0
 *  ITEM OWNS 3 > v[5]
 * 
 * ARMOR CLAUSES
 * -------------
 * 
 * ARMOR <not> OWNS <armor> <op> <value>
 *  This clause checks how many of a certain armor can be found in the party's 
 *  inventory.
 *  <armor> can be either a armor ID, a variable referencing the armor ID (using the
 *  v[n] format) or the armor name itself (with no blank spaces).
 *  <op> represents all conditional operators listed for VALUE COMPARISON.
 *  <value> can be either a number or a variable value (using the v[n] format).
 * 
 * Examples:
 *  ARMOR OWNS Cloth < 100
 *  ARMOR NOT OWNS v[4] > 0
 *  ARMOR OWNS 3 > v[5]
 * 
 * WEAPON CLAUSES
 * -------------
 * 
 * WEAPON <not> OWNS <weapon> <op> <value>
 *  This clause checks how many of a certain weapon can be found in the party's 
 *  inventory.
 *  <weapon> can be either a weapon ID, a variable referencing the weapon ID (using the
 *  v[n] format) or the weapon name itself (with no blank spaces).
 *  <op> represents all conditional operators listed for VALUE COMPARISON.
 *  <value> can be either a number or a variable value (using the v[n] format).
 * 
 * Examples:
 *  WEAPON OWNS SilverSword < 100
 *  WEAPON NOT OWNS v[4] > 0
 *  WEAPON OWNS 3 > v[5]
 * 
 * OTHER CLAUSES
 * -------------
 * 
 * BATTLE TURN <op> <value>
 *  This can be used to check how many battle turns it took for the battle to complete.
 *  This clause acts the same as Enemy Death Turn if drop evaluations are set to run
 *  at each enemy death instead of battle completion.
 *  <op> represents all conditional operators listed for VALUE COMPARISON.
 *  <value> can be either a number or a variable value (using the v[n] format).
 * 
 * Examples:
 *  BATTLE TURN < v[4]
 *  BATTLE TURN === 1
 * 
 * RANDOM <value>
 *  This clause can be used to consider a random chance of the condition being met.
 *  <value> can be a number followed by the % character to indicate percentage, a
 *  float number without the % sign (which will then be considered a percentage in
 *  decimal notation) or a variable referencing the probability of meeting the 
 *  condition (using the v[n] format, and limiting values to the interval of 0 to 100).
 * 
 * Examples:
 *  RANDOM 5%
 *  RANDOM v[5]
 *  RANDOM 0.5
 *  RANDOM 5.5%
 * 
 * EVAL <eval string>
 *  This clause can be used to evaluate a complex condition using actual javascript
 *  code. A few objects can be referenced for ease of use:
 *      'v[n]' to access the n variable value
 *      's[n]' to access the n switch value
 *      'e' to access Game_Enemy functions (in particular the enemy being assessed)
 *      'a' to access the $gameActors variable
 *      'p' to access the $gameParty variable
 *  This is an advanced clause that requires programming knowledge. Use it carefully.
 * 
 * Examples:
 *  EVAL v[1] >= p.battleMembers().length
 *  EVAL s[v[2]] === a.actor(v[4]).isAlive()
 * 
 * 
 * CLAUSE AGGREGATION
 * ------------------
 * 
 * The plugin also allows multiple clauses to be evaluated together when applying an
 * increase/decrease on drop chance. To use this feature enter two or more clauses
 * in the same line of text connecting them with the &&& character sequence. You can
 * find a few examples below:
 *  RANDOM 50% &&& BATTLE TURN < 5
 *      This condition will be met only if the random check beats the 50% odds AND
 *      if the battle has ended with less than 5 turns.
 *  ITEM OWNS Potion < 100 &&& ENEMY SKILL LAST Drain &&& RANDOM 0.8
 *      This condition will be met only if the party owns less the 100 Potions, the
 *      last skill used by the enemy is Drain and the random check beats the 80% odds
 *  PARTY GOLD > 1000 &&& ENEMY DEATH DAMAGE > 500
 *      This condition will be met only if the player have more than 1000 gold and
 *      the deathblow dealt to the enemy caused more than 500 damage.
 * 
 * 
 * RESERVED CHARACTERS
 * -------------------
 * 
 * To ensure the plugin work as intended, you cannot use : or &&& in any way but
 * as described in this documentation. The ':' use is reserved to separate condition
 * from effect from the stop clause, and '&&&' is used to group two or more clauses
 * in the same condition.
 * 
 * 
 * STOP CLAUSES
 * ------------
 * 
 * Stop clauses can be used to interrupt a series of conditional clauses if a condition
 * is or is not met. This is useful to chain conditional rules to one another, ensuring
 * that one specific check block all others and avoid unnecessary processing.
 * 
 * Here's an example using a sequence of conditional rules:
 *      RANDOM 10%: +15%
 *      ENEMY DEATH DAMAGE > 500: +10%
 *      ENEMY LEVEL > 10: +15% :stop if false
 *      ENEMY SKILL LAST 5: +0.05
 *          Breaking down this sequence:
 *              - If a randomly generated value between 0 and 100 is less than 10 (a 10%
 *                chance), drop chance increases by 15%;
 *              - If the deathblow to the enemy dealt more than 500 damage, drop chance
 *                increases another 10%;
 *              - If enemy level is higher than 10, drop chance increases another 15%.
 *                However, if the enemy's level is less or equal to 10, do not increase
 *                drop chances and do not test any remaining clauses;
 *              - If enemy level is higher than 10 and its last used skill was skill ID 5,
 *                increase drop chance by 5%;
 * 
 * =============================================================================
 * Enemy Note Tags
 * =============================================================================
 * 
 * Conditional drop rules can also be configured using enemy note tags. If you do use
 * note tags though, be aware that including at least one item rule via note tag will
 * cause the plugin to discard any clauses configured through the Plugin Manager or
 * JSON file, regardless of which item it affects.
 * 
 * To start a conditional drop note block, add the following pattern to your enemy
 * note box:
 *  <TAA_ED: <item type>=<item reference>, amount=<value>, attenuation=<type code>;<value>>
 *      <item type> is either 'item', 'armor' or 'weapon', and determine which type of
 *      item may be dropped after evaluating drop probabilities.
 *      <item reference> is an item, armor or weapon ID (depending on what <item type>
 *      is set to), a variable referencing an ID (using the v[n] format) or the dropped
 *      object name without blank spaces.
 *      <value> is an integer. When used with the 'amount' property it must be positive
 *      and higher than zero. For the attenuation settings it can assume a negative value.
 *      <type code> must be one of three values:
 *          0: Shared probability. That means that the drop chance is tested a single
 *            time for all items. If it pass, all items are dropped. If it fails,
 *            none are.
 *         1: Repeat probability. That means that each item has its own chance of
 *            being dropped, but the same probability is applied to each one of them.
 *         2: Attenuate probability. That means that for each item instance evaluated,
 *            drop probability for the next one is reduced/increased by the <attenuation 
 *            value>. This applies individually to items, armors and weapons. So if
 *            a drop rule is set to drop 2 of each, the conditional rules calculate
 *            the initial probability for the first item, armor and weapon, but the
 *            second of each one of them is attenuated.
 *      If the 'amount' property is not present, the plugin will assume the drop rules
 *      apply to a single object unit.
 *      Both 'amount' and 'attenuation' are optional parameters and can be omitted.
 * 
 * After entering the opening tag, insert your conditional clauses (one per line) and
 * close your conditional block with the tag </TAA_ED>
 * 
 * Here's a few examples:
 *  <TAA_ED: item=Potion>
 *  PARTY GOLD > 1000:+5%
 *  BATTLE TURN < v[1]: +10%
 *  RANDOM 15%: +20%
 *  </TAA_ED>
 * 
 *  <TAA_ED: armor=4, amount=2, attenuation=2;-50>
 *  PARTY NOT ARMOR 4: +10%
 *  v[5] >= 15: + 5%
 *  ENEMY DEATH DAMAGE > 500: +20%
 *  </TAA_ED>
 * 
 *  <TAA_ED: item=v[5], amount=1>
 *  s[1] is true: +10%
 *  v[6] < 100: +5%
 *  ENEMY STRUCK LAST BY SKILL 5: + 10%
 *  </TAA_ED>
 * 
 * ============================================================================
 * Default Parameters
 * ============================================================================
 * 
 * Default Drop Rules
 *  Use this structure to setup default conditional drop rules when none are present
 *  in drop configs. This will be applied to loot configured through the engine, so
 *  'Engine Settings' needs to be set to 'Keep'.
 * 
 * Engine Settings
 *  This parameter is used to determine how the plugin will handle default engine
 *  drop configurations. You can choose to either keep or discard. If set to 'Keep',
 *  the plugin will run the engine default drop settings and then add conditional
 *  drops to the resulting set. If set to 'Discard', engine drop settings are
 *  skipped, and only conditional drop rules are run.
 * 
 * Chance Processing
 *  Use this to determine when the conditional drop rules will be processed. It
 *  can happen either immediately after enemy death, or if it should wait the end
 *  of the battle.
 * 
 * Drop Rate Variable
 *  Select a variable which can be used to affect drop rate. If the variable value
 *  is 0, drop rate is not affected and the chance calculated based on conditional
 *  rules is used. If less than 0, the value is transformed into percentage and
 *  reduces the drop rate. If greater than 0, the value is turned into percentage
 *  and increases drop rate.
 * 
 * ============================================================================
 * Debugging tips
 * ============================================================================
 * 
 * By default, the plugin won't print any info to the console (this is the
 * recommended setup for a game deployment). However, during game development you
 * may want to check on things with a little more detail.
 * 
 * Inside the plugin file, right at the start of the file, you'll find a line
 * like follows:
 *  TAA.ed.trace = 0;
 * 
 * This controls the amount of log generated by the plugin. There's a bunch of
 * predefined log messages that are hidden from the console when it's set as 0.
 * Increasing it to 1, for example, will add a message stating how long it took
 * to process drop tags for each enemy. Increasing it to 2 will log details on
 * each clause processing. Keep in mind though, that the more messages the plugin
 * has to print to log, the more time it takes to go through all clauses. In
 * other words, increasing trace to 2 can seriously delay processing.
 * 
 * ============================================================================
 * Script Calls
 * ============================================================================
 * 
 * $gameSystem.edChanceProcessing()
 *  Returns 0 if chance processing is set to run at enemy death, or 1 if set to run
 *  at battle end.
 * 
 * $gameSystem.edUpdateChanceProcessing(option)
 *  Chance chance processing settings. option can be either 0 to run at enemy death,
 *  or 1 to run at battle end.
 * 
 * $gameSystem.edUseEngineSettings()
 *  Returns true if engine drops are kept, or false if they are set to be discarded.
 * 
 * $gameSystem.edEnableEngineSettings()
 *  Enables the Engine Settings parameter, so that engine drops are kept with 
 *  conditional drops.
 * 
 * $gameSystem.edDisableEngineSettings()
 *  Disables the Engine Settings parameter, so that engine drops are discarded and
 *  only conditional drops are considered.
 * 
 * ============================================================================
 * Plugin Commands (MV)
 * ============================================================================
 * 
 * EnemyDrops Chance Death
 * EnemyDrops Chance Battle
 *  Changes when conditional drop rules are processed. If 'Death', they are processed
 *  the moment an enemy dies (for that particular enemy). If 'Battle', all processing
 *  occurs at battle end.
 * 
 * EnemyDrops Engine Enable
 * EnemyDrops Engine Disable
 *  Enable or disable Engine Settings. If enabled, any enemy drops configured through
 *  the engine default drop configuration will run as set, and conditional drops are
 *  added to this results. If disabled, only conditional drops are considered.
 *
 * ============================================================================
 * Plugin Commands (MZ)
 * ============================================================================
 * 
 * Chance Processing Settings
 *  Manage when drop processing takes place. If can be set to run either at each
 *  individual enemy death, or all enemies together at battle end.
 * 
 * Engine Settings
 *  Manage how the plugin handles drops configured through the default engine
 *  configs. If enabled, any enemy drops configured through the engine will run
 *  as set, and conditional drops are added to this result. If disabled, only
 *  conditional drops are considered.
 *
 * ============================================================================
 * Changelog
 * ============================================================================
 *
 * Version 0.9.0:
 *  - Beta release;
 * Version 1.0.0:
 *  - Fixed a bug on default drop rules;
 *  - Refactored a bit of code to allow easier future plugins compatibility;
 * Version 1.0.1:
 *  - Fixed a bug that could cause game crash while preparing plugin indexes on
 * game boot;
 *  - Fixed a bug that would cause the max amount of drops of an item to always
 * be applied, ignoring the chance calculations;
 *  - Fixed a bug that would cause the amount of drops to be ignored, and never
 * drop more than one item. This scenario would happen only when using the Plugin
 * Manager as datasource;
 *
 * ============================================================================
 * End of Help
 * ============================================================================
 * 
 * 
 * 
 * =================================================================================
 * Commands (MZ)
 * =================================================================================
 * 
 * @command chanceProcessing
 * @text Chance Processing Settings
 * @desc Manage when drop chance processing takes place.
 * 
 * @arg process
 * @text Processing Type
 * @type select
 * @option At enemy death
 * @value 0
 * @option At battle end
 * @value 1
 * @default 0
 * @desc Determines when drop chance processing happens.
 * 
 * @command engineSettings
 * @text Engine Settings
 * @desc Manage how the plugin handles drops configured through the engine.
 * 
 * @arg Settings
 * @type boolean
 * @on Keep
 * @off Discard
 * @default false
 * @desc If set to keep (true), drops defined in the engine will pass regardless of drop rules. Otherwise they're discarded.
 * 
 * =================================================================================
 * Parameters
 * =================================================================================
 * 
 * @param ---DataSource Config---
 * @default ------------------------
 * 
 * @param SourceType
 * @parent ---DataSource Config---
 * @text DataSource Type
 * @type select
 * @option JSON File
 * @option Plugin Manager
 * @default JSON File
 * @desc Select the main source type for enemy drops data.
 * 
 * @param JSON Config
 * @parent ---DataSource Config---
 * @type struct<JsonConfig>
 * @desc Configure properties when using a JSON file as a datasource.
 * @default {"File":"EnemyExtra.json","ID Object":"id","Drops Array Object":"drops","Drop Type Object":"type","Item ID Object":"item","Armor ID Object":"armor","Weapon ID Object":"weapon","Base Chance Object":"base","Drop Amount Object":"amount","Iteration Effect Object":"iterationEffect","Iteration Type Object":"type","Iteration Attenuation Object":"attenuation","Conditions Array Object":"conditions"}
 * 
 * @param Plugin Manager Data
 * @parent ---DataSource Config---
 * @type struct<PMData>[]
 * @desc Configure enemy drops data through Plugin Manager if this datasource type is selected.
 * @default []
 * 
 * @param ---Default Settings---
 * @default ------------------------
 * 
 * @param Default Drop Rules
 * @parent ---Default Settings---
 * @type struct<DefaultDrops>
 * @default {"Base Chance":"0","Iteration Effect":"{\"Type\":\"0\",\"Attenuation\":\"0\"}","Drop Conditions":"[]"}
 * @desc Configure default drop rules, which will be applied to all enemies that don't have their own rules.
 * 
 * @param Engine Settings
 * @parent ---Default Settings---
 * @type boolean
 * @on Keep
 * @off Discard
 * @default false
 * @desc If set to keep (true), drops defined in the engine will pass regardless of drop rules. Otherwise they're discarded.
 * 
 * @param Chance Processing
 * @parent ---Default Settings---
 * @type select
 * @option At Enemy Death
 * @value 0
 * @option At Battle End
 * @value 1
 * @default 0
 * @desc Determines when drop chance processing happens.
 * 
 * @param Drop Rate Variable
 * @parent ---Default Settings---
 * @type variable
 * @default 
 * @desc Select a variable to use as a drop rate modifier.
 * 
 * @param ---Battle Messages---
 * @default ------------------------
 * 
 * @param Summarize Message
 * @parent ---Battle Messages---
 * @type boolean
 * @on Enable
 * @off Disable
 * @default true
 * @desc If enabled, each item will be listed only once. If disabled, the engine default format is used.
 * 
 * @param Drop Message Format
 * @parent ---Battle Messages---
 * @type text
 * @default %1 x%2
 * @desc If Summarize Message is enabled, this defines the item listing format. %1- item name. %2- item amount.
 * 
 */

//=============================================================================
// JSON File Configuration
//=============================================================================
 /*~struct~JsonConfig:
 * @param File
 * @type text
 * @default EnemyExtra.json
 * @desc Enemy drops JSON file on the data folder.
 * 
 * @param ID Object
 * @type text
 * @default id
 * @desc Object for the enemy ID (default: id)
 * 
 * @param Drops Array Object
 * @type text
 * @default drops
 * @desc Object referencing the drops array object.
 * 
 * @param Drop Type Object
 * @type text
 * @default type
 * @desc Object referencing the drop type (item, armor, weapon or all).
 * 
 * @param Item ID Object
 * @type text
 * @default item
 * @desc Object referencing the item to be dropped.
 * 
 * @param Armor ID Object
 * @type text
 * @default armor
 * @desc Object referencing the armor to be dropped.
 * 
 * @param Weapon ID Object
 * @type text
 * @default weapon
 * @desc Object referencing the weapon to be dropped.
 * 
 * @param Base Chance Object
 * @type text
 * @default base
 * @desc Object referencing the drop rate base probability.
 * 
 * @param Drop Amount Object
 * @type text
 * @default amount
 * @desc Object referencing the number of times the item will be dropped.
 * 
 * @param Iteration Effect Object
 * @type text
 * @default iterationEffect
 * @desc Object describing the iteration effect for when the rule applies to an amount of more than one of an item.
 * 
 * @param Iteration Type Object
 * @type text
 * @default type
 * @desc Object describing the iteration type.
 * 
 * @param Iteration Attenuation Object
 * @type text
 * @default attenuation
 * @desc Object describing the iteration attenuation.
 * 
 * @param Conditions Array Object
 * @type text
 * @default conditions
 * @desc Object referencing the conditions array, which allows for conditional probability.
 * 
 */

//=============================================================================
// Plugin Manager Data
//=============================================================================
 /*~struct~PMData:
 * @param Enemy ID
 * @type enemy
 * @default 1
 * @desc Enemy to apply drop conditional rules to.
 * 
 * @param Item Drops
 * @type struct<PMDrops>[]
 * @default []
 * @desc Drop settings array.
 * 
 */

 /*~struct~PMDrops:
 * @param Drop Type
 * @type select
 * @option Item
 * @option Armor
 * @option Weapon
 * @option All
 * @default Item
 * @desc Which drop are we setting up? Item, armor, weapon or all (if not blank)?
 *
 * @param Item
 * @type item
 * @default 
 * @desc Item to customize drop conditions.
 * 
 * @param Armor
 * @type armor
 * @default 
 * @desc Armor to customize drop conditions.
 * 
 * @param Weapon
 * @type weapon
 * @default 
 * @desc Weapon to customize drop conditions.
 * 
 * @param Base Chance
 * @type number
 * @min 0
 * @max 100
 * @decimals 2
 * @default 0
 * @desc Drop rate base probability (%).
 * 
 * @param Amount
 * @type number
 * @min 1
 * @max 999
 * @default 1
 * @desc Defines the amount to be dropped if rule succeeds.
 * 
 * @param Iteration Effect
 * @type struct<DropIteration>
 * @default {"Type":"0","Attenuation":"0"}
 * 
 * @param Drop Conditions
 * @type text[]
 * @default []
 * @desc Drop settings array.
 * 
 */

 /*~struct~DropIteration: 
 * @param Type
 * @type select
 * @option If drop succeeds, drop all (shared)
 * @value 0
 * @option Apply probability for each unit (repeat)
 * @value 1
 * @option Attenuate chance for each unit (attenuate)
 * @value 2
 * @default 0
 * @desc Defines how the plugin will handle drops of two or more instances of the same item (in the same rule).
 * 
 * @param Attenuation
 * @type number
 * @min -100
 * @max 100
 * @default 0
 * @desc If 'attenuate' type is selected, defines how to change probability for each additional unit (%).
 * 
 */

 /*~struct~DefaultDrops: 
 * @param Base Chance
 * @type number
 * @min 0
 * @max 100
 * @decimals 2
 * @default 0
 * @desc Drop rate base probability (%).
 * 
 * @param Iteration Effect
 * @type struct<DropIteration>
 * @default {"Type":"0","Attenuation":"0"}
 * 
 * @param Drop Conditions
 * @type text[]
 * @default []
 * @desc Drop settings array.
 * 
 */

//=============================================================================
// Local Functions
//=============================================================================

TAA.ed.functions = TAA.ed.functions || {};
TAA.ed.functions.log = function(){
    var args = Array.prototype.slice.call(arguments, 0);
    var trace = args[0];
    if(!isNaN(trace) && trace <= TAA.ed.trace){
        var msg = "";
        for(var i=1; i<args.length; i++){
            msg += args[i] + "\n";
        }
        console.log(msg);
    }
};

TAA.ed.functions.getProcessingSettings = function(param){
    if(!isNaN(param)) return 0;
    var value = parseInt(param);
    if(value !== 1) return 0;
    return value;
};

//=============================================================================
// Parameters Setup
//=============================================================================

TAA.ed.Parameters = TAA.ed.Parameters || {};
var Parameters = PluginManager.parameters(TAA.ed.PluginName);

// Datasource parameters
TAA.ed.Parameters.SourceType = Parameters['SourceType'] || 'JSON File';
TAA.ed.Parameters.JsonConfig = JSON.parse(Parameters['JSON Config']);
TAA.ed.Parameters.PMData = JSON.parse(Parameters['Plugin Manager Data']);

// Default settings
TAA.ed.Parameters.Default = TAA.ed.Parameters.Default || {};
TAA.ed.Parameters.Default.drops = Parameters['Default Drop Rules'] ? JSON.parse(Parameters['Default Drop Rules']) : {};
TAA.ed.Parameters.Default.EngineSettings = Parameters['Engine Settings'] === 'true';
TAA.ed.Parameters.Default.ChanceProcessing = TAA.ed.functions.getProcessingSettings(Parameters['Chance Processing']);
TAA.ed.Parameters.Default.DropRateVar = isNaN(Parameters['Drop Rate Variable']) ? 0 : parseInt(Parameters['Drop Rate Variable']);

// Battle Messages
TAA.ed.Parameters.BattleMsg = TAA.ed.Parameters.BattleMsg || {};
TAA.ed.Parameters.BattleMsg.Summarize = Parameters['Summarize Message'] === 'true';
TAA.ed.Parameters.BattleMsg.Format = Parameters['Drop Message Format'];

//=============================================================================
// DataManager
//=============================================================================

TAA.ed.alias.DataManager = TAA.ed.alias.DataManager || {};
TAA.ed.alias.DataManager.createGameObjects = DataManager.createGameObjects;
DataManager.createGameObjects = function(){
    TAA.ed.alias.DataManager.createGameObjects.call(this);
    if(!EnemyDataManager._initialized)
        EnemyDataManager.initialize();
};

DataManager.taaEnemyDropsDataVar = function(file){
    for(var i=DataManager._databaseFiles.length - 1; i>=0; i--){
        var tmp = DataManager._databaseFiles;
        if(tmp[i].name === '$taaEnemyData'){
            if(tmp[i].src !== file)
                return '$tempEnemyDrops';
            else
                break;
        }
    }
    return '$taaEnemyData';
};

if(TAA.ed.Parameters.SourceType === 'JSON File'){
    var file = TAA.ed.Parameters.JsonConfig['File'] || 'EnemyExtra.json';
    var dataVar = DataManager.taaEnemyDropsDataVar(file);
    DataManager._databaseFiles.push({ name: dataVar, src: file });
    if(dataVar !== '$taaEnemyData')
        var $tempEnemyDrops = $tempEnemyDrops || [];
    else
        var $tempEnemyDrops = null;

    TAA.ed.alias.DataManager.loadDataFile = DataManager.loadDataFile;
    DataManager.loadDataFile = function(name, src){
        var file = TAA.ed.Parameters.JsonConfig['File'] || 'EnemyExtra.json';
        file = 'Test_' + file;
        if(src === file)
            src = TAA.ed.Parameters.JsonConfig['File'] || 'EnemyExtra.json';
        TAA.ed.alias.DataManager.loadDataFile.call(this, name, src);
    }
}

var $taaEnemyData = $taaEnemyData || [];

//=============================================================================
// Enemy Data Manager
//=============================================================================

if(!EnemyDataManager){
    function EnemyDataManager() {
        throw new Error('This is a static class');
    };
}

/**
 * Initializes Enemy Data Manager
 * 
 * @static
 * @method initialize
 */
TAA.ed.alias.EnemyDataManager = TAA.ed.alias.EnemyDataManager || {};
TAA.ed.alias.EnemyDataManager.initialize = EnemyDataManager.initialize;
EnemyDataManager.initialize = function(){
    this._initialized = true;
    
    if(TAA.ed.alias.EnemyDataManager.initialize){
        TAA.ed.alias.EnemyDataManager.initialize.call(this);

        if(TAA.ed.Parameters.SourceType === 'JSON File'){
            this.initDropsByJson();
            if($tempEnemyDrops !== null){
                this.appendEnemyDropsData();
                $tempEnemyDrops = null;
            }
        }
        else {
            this.initDropsByPluginManager();
            this.organizePMData();
        }
    }
    else if(TAA.ed.Parameters.SourceType === 'JSON File'){
        this.initByJson();
        this.reorderJsonData();
    }
    else {
        this.initByPluginManager();
        this.organizePMData();
    }
    
    if(!this._indexesBuilt) this.buildNameIndexes();
};

/**
 * Initializes Enemy Data Manager through a JSON file
 * 
 * @static
 * @method initByJson
 */
TAA.ed.alias.EnemyDataManager.initByJson = EnemyDataManager.initByJson;
EnemyDataManager.initByJson = function(){
    if(this._initializedJson) return;
    if(TAA.ed.alias.EnemyDataManager.initByJson !== undefined)
        TAA.ed.alias.EnemyDataManager.initByJson.call(this);
    this.initDropsByJson();
    this._initializedJson = true;
};

/**
 * Initializes Enemy Data Manager through a JSON file
 * 
 * @static
 * @method initDropsByJson
 */
EnemyDataManager.initDropsByJson = function(){
    this._idDropsObject = TAA.ed.Parameters.JsonConfig['ID Object'] || "id";
    if(!this._idObject) this._idObject = this._idDropsObject;
    this._dropsObject = TAA.ed.Parameters.JsonConfig['Drops Object'] || "drops";
    this._dropTypeObject = TAA.ed.Parameters.JsonConfig['Drop Type Object'] || "type";
    this._itemIdObject = TAA.ed.Parameters.JsonConfig['Item ID Object'] || "item";
    this._armorIdObject = TAA.ed.Parameters.JsonConfig['Armor ID Object'] || "armor";
    this._weaponIdObject = TAA.ed.Parameters.JsonConfig['Weapon ID Object'] || "weapon";
    this._baseDropRateObject = TAA.ed.Parameters.JsonConfig['Base Chance Object'] || "base";
    this._amountObject = TAA.ed.Parameters.JsonConfig['Drop Amount Object'] || "amount";
    this._iterationEffectObject = TAA.ed.Parameters.JsonConfig['Iteration Effect Object'] || "iterationEffect";
    this._iterationTypeObject = TAA.ed.Parameters.JsonConfig['Iteration Type Object'] || "type";
    this._attenuationObject = TAA.ed.Parameters.JsonConfig['Iteration Attenuation Object'] || "attenuation";
    this._dropConditionsObject = TAA.ed.Parameters.JsonConfig['Conditions Array Object'] || "conditions";
};

/**
 * Initializes Enemy Data Manager through Plugin Manager data
 * 
 * @static
 * @method initByPluginManager
 */
TAA.ed.alias.EnemyDataManager.initByPluginManager = EnemyDataManager.initByPluginManager;
EnemyDataManager.initByPluginManager = function(){
    if(this._initializedPM) return;
    this._initializedPM = true;
    if(TAA.ed.alias.EnemyDataManager.initByPluginManager !== undefined)
        TAA.ed.alias.EnemyDataManager.initByPluginManager.call(this);

    this.initDropsByPluginManager();
};

/**
 * Initializes Enemy Data Manager through Plugin Manager data
 * 
 * @static
 * @method initDropsByPluginManager
 */
EnemyDataManager.initDropsByPluginManager = function(){
    this._idDropsObject = 'Enemy ID';
    if(!this._idObject) this._idObject = this._idDropsObject;
    this._dropsObject = 'Item Drops';
    this._dropTypeObject = 'Drop Type';
    this._itemIdObject = "Item";
    this._armorIdObject = "Armor";
    this._weaponIdObject = "Weapon";
    this._baseDropRateObject = "Base Chance";
    this._amountObject = "Amount";
    this._iterationEffectObject = "Iteration Effect";
    this._iterationTypeObject = "Type";
    this._attenuationObject = "Attenuation";
    this._dropConditionsObject = "Drop Conditions";
};

/**
 * Organizes Plugin Manager Data
 * 
 * @static
 * @method organizePMData
 */
TAA.ed.alias.EnemyDataManager.organizePMData = EnemyDataManager.organizePMData;
EnemyDataManager.organizePMData = function(){
    if(this._organizedPM) return;
    this._organizedPM = true;
    // If there's a previous plugin with this function set, preserve initial code
    if(TAA.ed.alias.EnemyDataManager.organizePMData !== undefined)
        TAA.ed.alias.EnemyDataManager.organizePMData.call(this);
    
    for(var i=0; i < TAA.ed.Parameters.PMData.length; i++){
        var enemyData = this.parseDropPMData(TAA.ed.Parameters.PMData[i]);
        if(enemyData && !isNaN(enemyData[this._idObject])){
            var id = enemyData[this._idDropsObject] || enemyData[this._idObject];
            $taaEnemyData[id] = enemyData;
        }
    }
};

/**
 * Auxiliary function to parse Plugin Manager Data
 * 
 * @static
 * @method parseDropPMData
 */
EnemyDataManager.parseDropPMData = function(dataInput){
    try{
        var data = JSON.parse(dataInput);
    } catch(e){
        console.error("TAA_EnemyDrops: Failed to parse plugin manager data.");
        return null;
    }
    var enemyData = {};
    if(!data[this._idDropsObject]){
        return null;
    }
    else{
        var id = data[this._idDropsObject];
        if($taaEnemyData[id])
            enemyData = $taaEnemyData[id];
        else
            enemyData[this._idObject] = parseInt(data[this._idDropsObject]);
    }
    
    if(data[this._dropsObject]){
        var dropsObj = JSON.parse(data[this._dropsObject]);
        for(var i=0; i<dropsObj.length;i++){
            if(!dropsObj[i])
                continue;
            var obj = JSON.parse(dropsObj[i]);
            enemyData[this._dropsObject] = enemyData[this._dropsObject] || [];
            var orgData = {};
            orgData[this._dropTypeObject] = obj[this._dropTypeObject];
            if(!isNaN(obj[this._itemIdObject]) && parseInt(obj[this._itemIdObject]) > 0)
                orgData[this._itemIdObject] = parseInt(obj[this._itemIdObject]);
            if(!isNaN(obj[this._armorIdObject]) && parseInt(obj[this._armorIdObject]) > 0){
                orgData[this._armorIdObject] = parseInt(obj[this._armorIdObject]);
            }
            if(!isNaN(obj[this._weaponIdObject]) && parseInt(obj[this._weaponIdObject]) > 0)
                orgData[this._weaponIdObject] = parseInt(obj[this._weaponIdObject]);
            var baseChance = isNaN(obj[this._baseDropRateObject]) ? 0 : parseFloat(obj[this._baseDropRateObject]);
            orgData[this._baseDropRateObject] = baseChance;
            if(!obj[this._amountObject] || isNaN(obj[this._amountObject]) || parseInt(obj[this._amountObject]) === 1)
                orgData[this._amountObject] = 1;
            else if(parseInt(obj[this._amountObject]) > 1){
				orgData[this._amountObject] = parseInt(obj[this._amountObject]);
                orgData[this._iterationEffectObject] = {};
                orgData[this._iterationEffectObject][this._iterationTypeObject] = 0;
                orgData[this._iterationEffectObject][this._attenuationObject] = 0;
                if(!obj[this._iterationEffectObject]){
                    if(!obj[this._iterationEffectObject][this._iterationTypeObject] && ['0','1','2'].contains(obj[this._iterationEffectObject][this._iterationTypeObject]))
                        orgData[this._iterationEffectObject][this._iterationTypeObject] = parseInt(obj[this._iterationEffectObject][this._iterationTypeObject]);
                
                    if(!obj[this._iterationEffectObject][this._attenuationObject] && !isNaN(obj[this._iterationEffectObject][this._attenuationObject])){
                        orgData[this._iterationEffectObject][this._attenuationObject] = parseInt(obj[this._iterationEffectObject][this._attenuationObject]);
                    }
                }
            }
            if(obj[this._dropConditionsObject])
                orgData[this._dropConditionsObject] = JSON.parse(obj[this._dropConditionsObject]);

            enemyData[this._dropsObject].push(orgData);
        }
    }
    return enemyData;
};

/**
 * Auxiliary function to reorder enemy data loaded through a JSON file
 * 
 * @static
 * @method reorderJsonData
 */
TAA.ed.alias.EnemyDataManager.reorderJsonData = EnemyDataManager.reorderJsonData;
EnemyDataManager.reorderJsonData = function(){
    if(TAA.ed.alias.EnemyDataManager.reorderJsonData)
        return TAA.ed.alias.EnemyDataManager.reorderJsonData.call(this);
    
    var orderedArray = [];
    if($taaEnemyData === undefined || $taaEnemyData === null || $taaEnemyData.length === 0)
        return;
    
    for(var i=0; i< $taaEnemyData.length; i++){
        if($taaEnemyData[i] !== undefined && $taaEnemyData[i][this._idDropsObject] !== undefined){
            var id = $taaEnemyData[i][this._idDropsObject];

            if(orderedArray[id] !== undefined) 
                console.error("Duplicated ID: " + id + ". Keeping the first entry.");
            else {
                var data = $taaEnemyData[i];
                orderedArray[id] = data;
            }
        }
    }
    $taaEnemyData = orderedArray.slice();
};

/**
 * Appends Enemy Drops data from file to a previously existing enemy data.
 * 
 * @static
 * @method appendEnemyDropsData
 */
EnemyDataManager.appendEnemyDropsData = function(){
    for(var i=0; i<$tempEnemyDrops.length; i++){
        var data = $tempEnemyDrops[i];
        if(!data || !data[this._idDropsObject])
            continue;
        var id = data[this._idDropsObject];
        if($taaEnemyData[id] === null || $taaEnemyData[id] === undefined)
            $taaEnemyData[id] = {};
        
        var keys = Object.keys(data);
        for(var j=0; j<keys.length; j++){
            if(keys[j] === this._idDropsObject)
                continue;
            $taaEnemyData[id][keys[j]] = data[keys[j]];
        }
    }
};

/**
 * Return a normalized object, for ease of use
 * 
 * @static
 * @method getNormalizedData
 */
TAA.ed.alias.EnemyDataManager.getNormalizedData = EnemyDataManager.getNormalizedData;
if(TAA.ed.alias.EnemyDataManager.getNormalizedData === undefined) {
    EnemyDataManager.getNormalizedData = function(enemyId){
        var rawData = this.getEnemyData(enemyId);
        return this.prepareNormalizedObject(rawData);
    };
};

/**
 * Returns enemy level data with no normalization
 * 
 * @static
 * @method getEnemyData
 */
TAA.ed.alias.EnemyDataManager.getEnemyData = EnemyDataManager.getEnemyData;
if(TAA.ed.alias.EnemyDataManager.getEnemyData === undefined){
    EnemyDataManager.getEnemyData = function(enemyId){
        if($taaEnemyData[enemyId] !== null && $taaEnemyData[enemyId] !== undefined) return $taaEnemyData[enemyId];
        if(this.getDefaultSettings) return this.getDefaultSettings(enemyId);
        else return undefined;
    };
};

/**
 * Prepares a normalized enemy data object
 * 
 * @static
 * @method prepareNormalizedObject
 */
TAA.ed.alias.EnemyDataManager.prepareNormalizedObject = EnemyDataManager.prepareNormalizedObject;
EnemyDataManager.prepareNormalizedObject = function(rawData){
    var data = {};
    if(TAA.ed.alias.EnemyDataManager.prepareNormalizedObject !== undefined)
        data = TAA.ed.alias.EnemyDataManager.prepareNormalizedObject.call(this, rawData);
    data['id'] = data['id'] || rawData[this._idDropsObject];
    data['drops'] = [];
    if(rawData[this._dropsObject]){
        for(var i=0; i<rawData[this._dropsObject].length; i++){
            var item = rawData[this._dropsObject][i];
            var dropObj = {};
            if(!item) continue;
            if(item[this._itemIdObject])
                dropObj.item = item[this._itemIdObject];
            if(item[this._armorIdObject])
                dropObj.armor = item[this._armorIdObject];
            if(item[this._weaponIdObject])
                dropObj.weapon = item[this._weaponIdObject];
            dropObj.type = "item";
            if(item[this._dropTypeObject])
                dropObj.type = item[this._dropTypeObject];
            dropObj.base = item[this._baseDropRateObject];
            dropObj.amount = 1;
            if(item[this._amountObject])
                dropObj.amount = item[this._amountObject];
            dropObj.iterationEffect = {
                type: 0,
                attenuation: 0
            };
            var tmp = item[this._iterationEffectObject];
            if(tmp){
                if(!isNaN(tmp[this._iterationTypeObject]))
                    dropObj.iterationEffect.type = parseInt(tmp[this._iterationTypeObject]);
                if(!isNaN(tmp[this._attenuationObject])){
                    var value = parseInt(tmp[this._attenuationObject]);
                    if(value < 0) value = Math.max(-100, value)/100;
                    else value = Math.min(100, value)/100;
                    dropObj.iterationEffect.attenuation = value;
                }
            }
            dropObj.conditions = item[this._dropConditionsObject];
            data.drops.push(dropObj);
        }
    }
    return data;
};

/**
 * Create name indexes to make it faster to lookup an ID through a name
 * 
 * @static
 * @method buildNameIndexes
 */
EnemyDataManager.buildNameIndexes = function(){
    this._indexesBuilt = this._indexesBuilt || false;
    this._itemIndex = this._itemIndex || {};
    this._armorIndex = this._armorIndex || {};
    this._weaponIndex = this._weaponIndex || {};
    this._actorIndex = this._actorIndex || {};
    this._classIndex = this._classIndex || {};
    this._stateIndex = this._stateIndex || {};
    this._skillIndex = this._skillIndex || {};
    this._armorTypeIndex = this._armorTypeIndex || {};
    this._weaponTypeIndex = this._weaponTypeIndex || {};
    this._skillTypeIndex = this._skillTypeIndex || {};
    this._elementIndex = this._elementIndex || {};

    // Item Index
    if(Object.keys(this._itemIndex).length === 0){
        for(var i=1; i<$dataItems.length; i++){
            var item = $dataItems[i];
            if(item && item.name && item.name !== "")
                this._itemIndex[item.name.replace(/\s/g, '')] = i;
        }
    }

    // Armor Index
    if(Object.keys(this._armorIndex).length === 0){
        for(var i=1; i<$dataArmors.length; i++){
            var item = $dataArmors[i];
            if(item && item.name && item.name !== "")
                this._armorIndex[item.name.replace(/\s/g, '')] = i;
        }
    }

    // Weapon Index
    if(Object.keys(this._weaponIndex.length === 0)){
        for(var i=1; i<$dataWeapons.length; i++){
            var item = $dataWeapons[i];
            if(item && item.name && item.name !== "")
                this._weaponIndex[item.name.replace(/\s/g, '')] = i;
        }
    }

    // Actor Index
    if(Object.keys(this._actorIndex).length === 0){
        for(var i=1; i<$dataActors.length; i++){
            var item = $dataActors[i];
            if(item && item.name && item.name !== "")
                this._actorIndex[item.name.replace(/\s/g, '')] = i;
        }
    }

    // Class Index
    if(Object.keys(this._classIndex).length === 0){
        for(var i=1; i<$dataClasses.length; i++){
            var item = $dataClasses[i];
            if(item && item.name && item.name !== "")
                this._classIndex[item.name.replace(/\s/g, '')] = i;
        }
    }

    // State Index
    if(Object.keys(this._stateIndex).length === 0){
        for(var i=1; i<$dataStates.length; i++){
            var item = $dataStates[i];
            if(item && item.name && item.name !== "")
                this._stateIndex[item.name.replace(/\s/g, '')] = i;
        }
    }

    // Skill Index
    if(Object.keys(this._skillIndex).length === 0){
        for(var i=1; i<$dataSkills.length; i++){
            var item = $dataSkills[i];
            if(item && item.name && item.name !== "")
                this._skillIndex[item.name.replace(/\s/g, '')] = i;
        }
    }

    // Armor Type Index
    if(Object.keys(this._armorTypeIndex).length === 0){
        var aTypes = $dataSystem.armorTypes;
        for(var i=1; i<aTypes.length; i++){
            var item = aTypes[i];
            if(item && item !== "")
                this._armorTypeIndex[item.replace(/\s/g, '')] = i;
        }
    }

    // Weapon Type Index
    if(Object.keys(this._weaponTypeIndex).length === 0){
        var wTypes = $dataSystem.weaponTypes;
        for(var i=1; i<wTypes.length; i++){
            var item = wTypes[i];
            if(item && item !== "")
                this._weaponTypeIndex[item.replace(/\s/g, '')] = i;
        }
    }

    // Skill Type Index
    if(Object.keys(this._skillTypeIndex).length === 0){
        var sType = $dataSystem.skillTypes;
        for(var i=1; i<sType.length; i++){
            var item = sType[i];
            if(item && item !== "")
                this._skillTypeIndex[item.replace(/\s/g, '')] = i;
        }
    }

    // Element Index
    if(Object.keys(this._elementIndex).length === 0){
        var element = $dataSystem.elements;
        for(var i=1; i<element.length; i++){
            var item = element[i];
            if(item && item !== "")
                this._elementIndex[item.replace(/\s/g, '')] = i;
        }
    }

    this._indexesBuilt = true;
};

/**
 * Parses a text and returns a variable value if it follows the pattern
 * 
 * @static
 * @method getVarValueFromText
 */
EnemyDataManager.getVarValueFromText = function(str, nameIndex){
    if(!isNaN(str)){
        return parseInt(str);
    }
    else if(str && str.match(/v\[(\d+)\]/i)){
        return $gameVariables.value(RegExp.$1);
    }
    else if(nameIndex && nameIndex[str]){
        return nameIndex[str];
    }
    return 0;
};

/**
 * Runs an eval for item drop rule analysis
 * 
 * @static
 * @method evalDropComparison
 */
EnemyDataManager.evalDropComparison = function(evalStr, enemy, actor){
    var v = $gameVariables._data;
    var s = $gameSwitches._data;
    var e = enemy;
    var a = $gameActors;
    var p = $gameParty;
    TAA.ed.functions.log(3, "Evaluating drop string: " + evalStr);
    try{
        result = eval(evalStr);
    } catch(e){
        console.error("TAA_EnemyDrops: Failed to eval '" + evalStr + "'. Assuming the result is FALSE.");
        result = false;
    }
    return result;
};

/**
 * Parse effect settings into a normalized object
 * 
 * @static
 * @method parseEffectValue
 */
EnemyDataManager.parseEffectValue = function(effectStr){
    if(!effectStr) return;
    var obj = {};
    if(effectStr.match(/(\+|\-|\*|\=)\s*(\d+(?:\.\d+)?|v\[\d+\])\s*(\%)?/)){
        var op = RegExp.$1;
        var value = RegExp.$2;
        var type = RegExp.$3;
        if(isNaN(value)){
            if(value.match(/v\[(\d+)\]/i))
                value = $gameVariables.value(RegExp.$1);
            else
                value = 0;
        }
        if(type)
            value = parseFloat(value)/100;
        else
            value = parseFloat(value);
        
        obj = {
            op: op,
            value: value
        };
    }
    return obj;
};

/**
 * Parse drop rules related to actor values
 * 
 * @static
 * @method parseActorDropRule
 */
EnemyDataManager.parseActorDropRule = function(ruleArray, clause){
    var id = this.getVarValueFromText(ruleArray[1], this._actorIndex);
    if(id <= 0) return false;
    var actor = $gameActors.actor(id);
    if(!actor) return false;
    var evalStr = "";
    var evalResult = false;
    var index = 2;
    var not = false;

    if(ruleArray[2].toUpperCase() === 'NOT'){
        index = 3;
        not = true;
    }

    if(ruleArray[index].toUpperCase() === 'IS'){
        evalStr = ruleArray[index+1].toUpperCase()
        if(evalStr === 'ALIVE'){
            evalResult = actor.isAlive();
        }
        else if(evalStr === 'DEAD'){
            evalResult = actor.isDead();
        }
    }
    else if(ruleArray[index].toUpperCase() === 'IN' && ruleArray[index+1].toUpperCase() === 'PARTY'){
        evalStr = "in party";
        evalResult = $gameParty.battleMembers().contains(actor);
    }
    else if(ruleArray[index].toUpperCase() === 'STATE'){
        var stateId = this.getVarValueFromText(ruleArray[index+1], this._stateIndex);
        evalStr = "actor.isStateAffected(" + stateId + ")";
        if(stateId > 0)
            evalResult = this.evalDropComparison("actor.isStateAffected(" + stateId + ")", undefined, actor);
    }
    else if(ruleArray[index].toUpperCase() === 'CLASS'){
        var classId = this.getVarValueFromText(ruleArray[index+1], this._classIndex);
        evalStr = "(actor._classId [" + actor._classId + "] === " + classId + ")";
        if(classId > 0)
            evalResult = (actor._classId === classId);
    }
    else if(ruleArray[index].toUpperCase() === 'ARMOR'){
        var type = false;
        evalStr = "actor has armor ";
        var armorId = 0;
        if(ruleArray[index+1].toUpperCase() === "TYPE"){
            type = true;
            index++;
            evalStr = "actor has armor type ";
            armorId = this.getVarValueFromText(ruleArray[index+2], this._armorTypeIndex);
        }
        else
            armorId = this.getVarValueFromText(ruleArray[index+1], this._armorIndex);
        index++;
        evalStr += armorId;
        if(armorId > 0){
            var armors = actor.armors();
            for(var j=0; j<armors.length; j++){
                if((type && armors[j].atypeId === armorId) || (!type && armors[j].id === armorId)) {
                    evalResult = true;
                    break;
                }
            }
        }
    }
    else if(ruleArray[index].toUpperCase() === 'WEAPON'){
        var type = false;
        var weaponId = 0;
        evalStr = "actor has weapon ";
        index++;
        if(ruleArray[index+1].toUpperCase() === "TYPE"){
            type = true;
            index++;
            evalStr = "actor has weapon type ";
            weaponId = this.getVarValueFromText(ruleArray[index], this._weaponTypeIndex);
        }
        else
            this.getVarValueFromText(ruleArray[index], this._weaponIndex);
        evalStr += weaponId;
        var weapons = actor.weapons();
        for(var j=0; j<weapons.length; j++){
            if((type && weapons[j].wtypeId === weaponId) || (!type && weapons[j].id === weaponId)) {
                evalResult = true;
                break;
            }
        }
    }
    else if(ruleArray[index].toUpperCase() === 'SKILL'){
        var skillId = this.getVarValueFromText(ruleArray[index+1], this._skillIndex);
        var level = undefined;
        if(TAA.skm && ruleArray[index+2].toUpperCase() === 'LEVEL'){
            level = this.getVarValueFromText(ruleArray[index+3]);
        }
        evalStr = "actor.hasSkill(" + skillId + ", " + level + ")";
        evalResult = actor.hasSkill(skillId, level);
    }
    else if(ruleArray[index].toUpperCase() === 'LEVEL'){
        var level = this.getVarValueFromText(ruleArray[index+2]);
        var op = ruleArray[index+1];

        if(level > 0){
            evalStr = actor.level + " " + op + " " + level;
            evalResult = this.evalDropComparison(evalStr);
        }
    }

    if(not)
        evalResult = !evalResult;
    
    TAA.ed.functions.log(2, "ACTOR CLAUSE\n--------------------", "evalStr: " + evalStr, "not: " + not, "effect: " + clause[1], "isStop: " + clause[2], "evalResult: " + evalResult);
    
    return evalResult;
};

/**
 * Parse drop rules related to party values
 * 
 * @static
 * @method parsePartyDropRule
 */
EnemyDataManager.parsePartyDropRule = function(ruleArray, clause){
    var members = $gameParty.battleMembers();
    var evalStr = "";
    var evalResult = false;
    var index = 1;
    var not = false;

    if(ruleArray[1].toUpperCase() === 'NOT'){
        index = 2;
        not = true;
    }

    if(ruleArray[index].toUpperCase() === 'STATE'){
        var stateId = this.getVarValueFromText(ruleArray[index+1], this._stateIndex);
        evalStr = "$gameParty.partyActiveStates().contains(" + stateId + ")";
        if(stateId > 0){
            evalResult = $gameParty.partyActiveStates().contains(stateId);
        }
    }
    else if(ruleArray[index].toUpperCase() === 'CLASS'){
        var classId = this.getVarValueFromText(ruleArray[index+1], this._classIndex);
        evalStr = "has any member in class " + classId;
        if(classId > 0){
            for(var i=0; i<members.length; i++){
                if(members[i]._classId === classId){
                    evalResult = true;
                    break;
                }
            }
        }
    }
    else if(ruleArray[index].toUpperCase() === 'ARMOR'){
        var type = false;
        evalStr = "has any member with armor ";
        var armorId = 0;
        index++;
        if(ruleArray[index].toUpperCase() === "TYPE"){
            type = true;
            index++;
            evalStr = "has any member with armor type ";
            armorId = this.getVarValueFromText(ruleArray[index], this._armorTypeIndex);
        }
        armorId = this.getVarValueFromText(ruleArray[index], this._armorIndex);
        evalStr += armorId;
        if(armorId > 0){
            for(var i=0; i<members.length; i++){
                var armors = members[i].armors();
                for(var j=0; j<armors.length; j++){
                    if((type && armors[j].atypeId === armorId) || (!type && armors[j].id === armorId)) {
                        evalResult = true;
                        break;
                    }
                }
                if(evalResult) break;
            }
        }
    }
    else if(ruleArray[index].toUpperCase() === 'WEAPON'){
        var type = false;
        evalStr = "has any member with weapon ";
        var weaponId = 0;
        index++;
        if(ruleArray[index].toUpperCase() === "TYPE"){
            type = true;
            index++;
            evalStr = "has any member with weapon type ";
            weaponId = this.getVarValueFromText(ruleArray[index], this._weaponTypeIndex);
        }
        else    
            weaponId = this.getVarValueFromText(ruleArray[index], this._weaponIndex);
        evalStr += weaponId;
        if(weaponId > 0){
            for(var i=0; i<members.length; i++){
                var weapons = members[i].weapons();
                for(var j=0; j<weapons.length; j++){
                    if((type && weapons[j].wtypeId === weaponId) || (!type && weapons[j].id === weaponId)) {
                        evalResult = true;
                        break;
                    }
                }
                if(evalResult) break;
            }
        }
    }
    else if(ruleArray[index].toUpperCase() === 'SKILL'){
        var skillId = this.getVarValueFromText(ruleArray[index+1], this._skillIndex);
        var level = undefined;
        if(TAA.skm && ruleArray[index+2].toUpperCase() === 'LEVEL'){
            level = this.getVarValueFromText(ruleArray[index+3]);
        }
        evalStr = "has any member with skill " + skillId + " (with level " + level + ")";
        for(var i=0; i<members.length; i++){
            if(members[i].hasSkill(skillId, level)){
                evalResult = true;
                break;
            }
        }
    }
    else if(ruleArray[index].toUpperCase() === 'LEVEL'){
        var type = 0;
        if(ruleArray[index+1].match(/(avg|max|min)/i)){
            switch((RegExp.$1).toUpperCase()){
                case 'AVG':
                    type = 1;
                    break;
                case 'MAX':
                    type = 2;
                    break;
                case 'MIN':
                    type = 3;
                    break;
            }
            index++;
        }
        var level = this.getVarValueFromText(ruleArray[index+2]);
        var op = ruleArray[index+1];
        
        if(type === 1)
            evalStr = $gameParty.averageBattlerLevel() + " " + op + " " + level;
        else if(type === 2 || (type === 0 && ['>', '>='].contains(op)))
            evalStr = $gameParty.highestBattlerLevel() + " " + op + " " + level;
        else if(type === 3 || (type === 0 && ['<', '<='].contains(op)))
            evalStr = $gameParty.lowestBattleLevel() + " " + op + " " + level;
        else {
            for(var i=0; i<members.length; i++){
                if(evalStr.length > 0)
                    evalStr += " || ";
                evalStr += members[i].level() + " " + op + " " + level;
            }
        }

        evalResult = this.evalDropComparison(evalStr);
    }
    else if(ruleArray[index].toUpperCase() === 'MEMBERS'){
        var value = this.getVarValueFromText(ruleArray[index+2]);
        var op = ruleArray[index+1];

        evalStr = $gameParty.members().length + " " + op + " " + value;
        evalResult = this.evalDropComparison(evalStr);
    }
    else if(ruleArray[index+1].toUpperCase() === 'MEMBERS'){
        var value = this.getVarValueFromText(ruleArray[index+3]);
        var op = ruleArray[index+2];

        var checkValue = undefined;
        switch(ruleArray[index].toUpperCase()){
            case 'ALIVE':
                checkValue = $gameParty.aliveMembers().length;
                break;
            case 'DEAD':
                checkValue = $gameParty.deadMembers().length;
                break;
            case 'BATTLE':
                checkValue = members.length;
                break;
        }
        if(checkValue === undefined){
            evalStr = "checkValue is undefined: " + ruleArray[index] + " Members";
            evalResult = false;
        }
        else{
            evalStr = checkValue + " " + op + " " + value;
            evalResult = this.evalDropComparison(evalStr);
        }
    }
    else if(ruleArray[index].toUpperCase() === 'GOLD'){
        var value = this.getVarValueFromText(ruleArray[index+2]);
        var op = ruleArray[index+1];

        evalStr = $gameParty.gold() + " " + op + " " + value;
        evalResult = this.evalDropComparison(evalStr);
    }

    if(not)
        evalResult = !evalResult;
    
    TAA.ed.functions.log(2, "PARTY CLAUSE\n--------------------", "evalStr: " + evalStr, "not: " + not, "effect: " + clause[1], "isStop: " + clause[2], "evalResult: " + evalResult);
    
    return evalResult;
};

/**
 * Parse drop rules related to enemy values
 * 
 * @static
 * @method parseEnemyDropRule
 */
EnemyDataManager.parseEnemyDropRule = function(ruleArray, clause, enemy){
    var evalStr = "";
    var evalResult = false;
    var index = 1;
    var not = false;

    if(ruleArray[1].toUpperCase() === 'NOT'){
        index = 2;
        not = true;
    }

    if(ruleArray[index].toUpperCase() === 'DEATH'){
        if(ruleArray[index+1].toUpperCase() === 'TURN'){
            var op = ruleArray[index+2];
            var value = this.getVarValueFromText(ruleArray[index+3]);
            evalStr = enemy.edDeathTurn() + " " + op + " " + value;
            evalResult = this.evalDropComparison(evalStr);
        }
        else if(['DAMAGE', 'DMG'].contains(ruleArray[index+1].toUpperCase())){
            var op = ruleArray[index+2];
            var value = this.getVarValueFromText(ruleArray[index+3]);
            evalStr = enemy.edLastDamage() + " " + op + " " + value;
            evalResult = this.evalDropComparison(evalStr);
        }
    }
    else if(TAA.el && ruleArray[index].toUpperCase() === 'LEVEL'){
        var op = ruleArray[index+1];
        var value = ruleArray[index+2];
        evalStr = enemy.getLevel() + " " + op + " " + value;
        evalResult = this.evalDropComparison(evalStr);
    }
    else if(ruleArray[index].toUpperCase() === 'STRUCK'){
        var type = 0;
        if(ruleArray[index+1].toUpperCase() === 'LAST'){
            type = 1;
        }
        else if(ruleArray[index+1].toUpperCase() === 'COUNT'){
            type = 2;
        }
        if(type > 0){
            index += 3;
            var value1 = 0;
            var hitValue = undefined;
            if(ruleArray[index].toUpperCase() === 'SKILL'){
                value1 = this.getVarValueFromText(ruleArray[index+1], this._skillIndex);
                hitValue = enemy.edLastSkillHit();
                countValue = enemy.edSkillHitCount(value1);
            }
            else if(ruleArray[index].toUpperCase() === 'ACTOR'){
                value1 = this.getVarValueFromText(ruleArray[index+1], this._actorIndex);
                hitValue = enemy.edLastActorHit();
                countValue = enemy.edActorHitCount(value1);
            }
            else if(ruleArray[index].toUpperCase() === 'ELEMENT'){
                value1 = this.getVarValueFromText(ruleArray[index+1], this._elementIndex);
                hitValue = enemy.edLastElementHit();
                countValue = enemy.edElementHitCount(value1);
            }
            else if(ruleArray[index].toUpperCase() === 'STYPE'){
                value1 = this.getVarValueFromText(ruleArray[index+1], this._skillTypeIndex);
                hitValue = enemy.edLastSkillTypeHit();
                countValue = enemy.edSkillTypeHitCount(value1);
            }
            else if(ruleArray[index].toUpperCase() === 'ITEM'){
                value1 = this.getVarValueFromText(ruleArray[index+1], this._itemIndex);
                hitValue = enemy.edLastItemHit();
                countValue = enemy.edItemHitCount(value1);
            }

            if(type === 1 && !isNaN(hitValue)){
                evalStr = "(" + value1 + " === " + hitValue + " && " + value1 + " > 0)";
                evalResult = (value1 === hitValue && value1 > 0);
            }
            else if(type === 2 && !isNaN(countValue)){
                var op = ruleArray[index+2];
                var value2 = this.getVarValueFromText(ruleArray[index+3]);
                evalStr = countValue + " " + op + " " + value2;
                evalResult = this.evalDropComparison(evalStr);
            }
        }
    }
    else if(ruleArray[index].toUpperCase() === 'SKILL'){
        var value1 = this.getVarValueFromText(ruleArray[index+2], this._skillIndex);
        if(ruleArray[index+1].toUpperCase() === 'LAST'){
            evalStr = "(" + value1 + " === " + enemy.edLastSkillUsed() + " && " + value1 + " > 0)";
            evalResult = (value1 === enemy.edLastSkillUsed() && value1 > 0);
        }
        else if(ruleArray[index+1].toUpperCase() === 'COUNT'){
            var op = ruleArray[index+2];
            var value2 = this.getVarValueFromText(ruleArray[index+3]);
            evalStr = enemy.edSkillUsedCount(value1) + " " + op + " " + value2;
            evalResult = this.evalDropComparison(evalStr);
        }
    }

    if(not)
        evalResult = !evalResult;
    
    TAA.ed.functions.log(2, "ENEMY CLAUSE\n--------------------", "evalStr: " + evalStr, "not: " + not, "effect: " + clause[1], "isStop: " + clause[2], "evalResult: " + evalResult);
    
    return evalResult;
};

/**
 * Parse drop rules related to party item usage
 * 
 * @static
 * @method parseItemDropRule
 */
EnemyDataManager.parseItemDropRule = function(ruleArray, clause){
    var id = this.getVarValueFromText(ruleArray[1], this._itemIndex);
    if(id <= 0) return false;
    var evalStr = "";
    var evalResult = false;
    var index = 2;
    var not = false;

    if(ruleArray[2].toUpperCase() === 'NOT'){
        index = 3;
        not = true;
    }

    if(ruleArray[index].toUpperCase() === 'LAST'){
        evalStr = id + " === " + $gameParty.edLastItemUsed();
        evalResult = (id === $gameParty.edLastItemUsed());
    }
    else if(ruleArray[index].toUpperCase() === 'TIMES'){
        var op = ruleArray[index+2];
        var value2 = this.getVarValueFromText(ruleArray[index+3]);
        evalStr = $gameParty.edItemUsedCount(id) + " " + op + " " + value2;
        evalResult = this.evalDropComparison(evalStr);
    }
    else if(ruleArray[index].toUpperCase() === 'OWNS'){
        var item = this.getVarValueFromText(ruleArray[index-1], this._itemIndex);
        var op = ruleArray[index+1];
        var count = this.getVarValueFromText(ruleArray[index+2]);
        evalStr = $gameParty.numItems($dataItems[item]) + " " + op + " " + count;
        evalResult = this.evalDropComparison(evalStr);
    }

    if(not)
        evalResult = !evalResult;
    
    TAA.ed.functions.log(2, "ITEM CLAUSE\n--------------------", "evalStr: " + evalStr, "not: " + not, "effect: " + clause[1], "isStop: " + clause[2], "evalResult: " + evalResult);
    
    return evalResult;
};

/**
 * Parse drop rules related to party armor collection
 * 
 * @static
 * @method parseArmorDropRule
 */
EnemyDataManager.parseArmorDropRule = function(ruleArray, clause){
    var id = this.getVarValueFromText(ruleArray[1], this._armorIndex);
    if(id <= 0) return false;
    var evalStr = "";
    var evalResult = false;
    var index = 2;
    var not = false;

    if(ruleArray[index].toUpperCase() === 'NOT'){
        index++;
        not = true;
    }

    if(ruleArray[index].toUpperCase() === 'OWNS'){
        var armor = this.getVarValueFromText(ruleArray[index-1], this._armorIndex);
        var op = ruleArray[index+1];
        var count = this.getVarValueFromText(ruleArray[index+2]);
        evalStr = $gameParty.numItems($dataArmors[armor]) + " " + op + " " + count;
        evalResult = this.evalDropComparison(evalStr);
    }

    if(not)
        evalResult = !evalResult;
    
    TAA.ed.functions.log(2, "ARMOR CLAUSE\n--------------------", "evalStr: " + evalStr, "not: " + not, "effect: " + clause[1], "isStop: " + clause[2], "evalResult: " + evalResult);
    
    return evalResult;
};

/**
 * Parse drop rules related to party weapon collection
 * 
 * @static
 * @method parseWeaponDropRule
 */
EnemyDataManager.parseWeaponDropRule = function(ruleArray, clause){
    var id = this.getVarValueFromText(ruleArray[1], this._weaponIndex);
    if(id <= 0) return false;
    var evalStr = "";
    var evalResult = false;
    var index = 2;
    var not = false;

    if(ruleArray[2].toUpperCase() === 'NOT'){
        index = 3;
        not = true;
    }

    if(ruleArray[index].toUpperCase() === 'OWNS'){
        var weapon = this.getVarValueFromText(ruleArray[index-1], this._weaponIndex);
        var op = ruleArray[index+1];
        var count = this.getVarValueFromText(ruleArray[index+2]);
        evalStr = $gameParty.numItems($dataWeapons[weapon]) + " " + op + " " + count;
        evalResult = this.evalDropComparison(evalStr);
    }

    if(not)
        evalResult = !evalResult;
    
    TAA.ed.functions.log(2, "WEAPON CLAUSE\n--------------------", "evalStr: " + evalStr, "not: " + not, "effect: " + clause[1], "isStop: " + clause[2], "evalResult: " + evalResult);
    
    return evalResult;
};

/**
 * Parse an evaluate drop rules
 * 
 * @static
 * @method parseDropRule
 */
EnemyDataManager.parseDropRule = function(clause, rule, ruleArray, enemy){
    var evalResult = false;
    if(!isNaN(ruleArray[0]) || ruleArray[0].match(/v\[\d+\]/i)){
        if(!['<', '<=', '>', '>=', '===', '!==', '==', '!='].contains(ruleArray[1])) evalResult = false;
        else{
            var value1 = this.getVarValueFromText(ruleArray[0]);
            var value2 = this.getVarValueFromText(ruleArray[2]);
            var evalStr = value1 + " " + ruleArray[1] + " " + value2;
            var evalResult = this.evalDropComparison(evalStr);
        }
        TAA.ed.functions.log(2, "VALUE COMPARISON\n--------------------", "evalStr: " + evalStr, "effect: " + clause[1], "isStop: " + clause[2], "evalResult: " + evalResult);
    }
    else if(rule.match(/(?:s|sw|switch)\s+\[(\d+)\]\s+(is|is\s+not)\s+(on|off|true|false)/i)){
        var sw = RegExp.$1;
        var op = RegExp.$2;
        var value = ['on', 'true'].contains((RegExp.$3).toLowerCase()) ? true : false;
        if(op.toLowerCase() === 'is not') value = !value;
        TAA.ed.functions.log(2, "SWITCH VALUE\n--------------------", "switchId: " + sw, "op: " + op, "value: " + value);
        if(!isNaN(sw) && $gameSwitches.value(sw) === value){
            evalResult = true;
        }
    }
    else if(rule.match(/(?:s|sw|switch)\s*\[(\d+)\]\s*(===|!==)\s*(?:s|sw|switch\[(\d+)\])/i)){
        var sw1 = RegExp.$1;
        var op = RegExp.$2;
        var sw2 = RegExp.$3;
        evalStr = "s[" + sw1 + "] " + op + "s[" + sw2 + "]";
        evalResult = this.evalDropComparison(evalStr);
        TAA.ed.functions.log(2, "SWITCH COMPARISON\n--------------------", "evalStr: " + evalStr, "effect: " + clause[1], "isStop: " + clause[2], "evalResult: " + evalResult);
    }
    else if(ruleArray[0].toUpperCase() === 'ACTOR'){
        evalResult = this.parseActorDropRule(ruleArray, clause);
    }
    else if(ruleArray[0].toUpperCase() === 'PARTY'){
        evalResult = this.parsePartyDropRule(ruleArray, clause);
    }
    else if(ruleArray[0].toUpperCase() === 'BATTLE'){
        if(['TURN', 'TURNS'].contains(ruleArray[1].toUpperCase())){
            var value = this.getVarValueFromText(ruleArray[3]);
            var op = ruleArray[2];
            evalStr = $gameTroop.turnCount() + " " + op + " " + value;
            evalResult = this.evalDropComparison(evalStr);
            TAA.ed.functions.log(2, "BATTLE TURN\n--------------------", "evalStr: " + evalStr, "effect: " + clause[1], "isStop: " + clause[2], "evalResult: " + evalResult);
        }
    }
    else if(ruleArray[0].toUpperCase() === 'ENEMY'){
        evalResult = this.parseEnemyDropRule(ruleArray, clause, enemy);
    }
    else if(ruleArray[0].toUpperCase() === 'ITEM'){
        evalResult = this.parseItemDropRule(ruleArray, clause);
    }
    else if(ruleArray[0].toUpperCase() === 'ARMOR'){
        evalResult = this.parseArmorDropRule(ruleArray, clause);
    }
    else if(ruleArray[0].toUpperCase() === 'WEAPON'){
        evalResult = this.parseWeaponDropRule(ruleArray, clause);
    }
    else if(ruleArray[0].toUpperCase() === 'RANDOM'){
        if(ruleArray[1].match(/(\d+(?:\.\d+)?)(\%)?/)){
            var num = parseFloat(RegExp.$1);
            var percent = RegExp.$2;
            if(percent) num = Number((num/100).toFixed(4));
            var chance = Math.random();
            evalStr = num + " >= " + chance;
            evalResult = num >= chance;
            TAA.ed.functions.log(2, "RANDOM CLAUSE\n--------------------", "evalStr: " + evalStr, "effect: " + clause[1], "isStop: " + clause[2], "evalResult: " + evalResult);
        }
    }
    else if(ruleArray[0].toUpperCase() === 'EVAL'){
        if(rule.match(/eval (.*)/i)){
            evalStr = RegExp.$1;
            evalResult = this.evalDropComparison(evalStr, enemy);
            TAA.ed.functions.log(2, "EVAL CLAUSE\n--------------------", "evalStr: " + evalStr, "effect: " + clause[1], "isStop: " + clause[2], "evalResult: " + evalResult);
        }
    }
    return evalResult;
};

/**
 * Parse drop clause to evaluate rule testing and return rule effects object
 * 
 * @static
 * @method parseDropClause
 */
EnemyDataManager.parseDropClause = function(str, enemy){
    var result = {
        value: 0,
        op:undefined,
        stop:false
    };
    if(!str || str === "") return result;
    var clause = str.trim().split(/:/);
    var rules = clause[0].split(/&&&/);
    var pass = false;
    if(!clause || clause.length < 2) return result;

    var effectObj = this.parseEffectValue(clause[1]);
    if(!effectObj){
        return result;
    }
    
    for(var index = 0; index < rules.length; index++){
        var test = rules[index].trim().split(/\s+/);
        if(!test || test.length < 2){
            pass = true;
            continue;
        }
        TAA.ed.functions.log(3, "CLAUSE: " + rules[index]);
        pass = this.parseDropRule(clause, rules[index], test, enemy);
        if(!pass)
            break;
    }

    if(pass){
        result.op = effectObj.op;
        result.value = effectObj.value;
    }
    // If stop clause is present
    if(clause[2]){
        var reg = new RegExp("stop\\s+if\\s+" + pass, "i");
        if(clause[2].match(reg))
            result.stop = true;
    }

    return result;
};

/**
 * Directs drop rule parsing to determine drop probability
 * 
 * @static
 * @method processItemDropRules
 */
EnemyDataManager.processItemDropRules = function(enemy, dropObj){
    if(!dropObj) return false;
    var chance = Math.max(0, dropObj.base)/100;
    var array = dropObj.conditions;
    for(var i=0; i<array.length; i++){
        var result = this.parseDropClause(array[i], enemy);
        var previousChance = chance;
        if(result && result.op && !isNaN(result.value)){
            switch(result.op){
                case '+':
                    chance += result.value;
                    break;
                case '-':
                    chance -= result.value;
                    break;
                case '*':
                    chance *= result.value;
                    break;
                case '=':
                    chance = result.value;
                    break;
                default:
                    console.error("####### no operator found #######");
                    chance = result.value;
            }
        }
        TAA.ed.functions.log(1, "--------------------", "Updated chance value: " + previousChance + " ----> " + chance, "--------------------");
        if(result.stop)
            break;
    }
    TAA.ed.functions.log(1, "## Calculated drop chance: " + chance);
    return chance;
};

/**
 * Determines how many instances of an item will be dropped according to calculated probability
 * and attenuation settings.
 * 
 * @static
 * @method applyDropAttenuation
 */
EnemyDataManager.applyDropAttenuation = function(chance, dropObj){
    var amount = (!isNaN(dropObj.amount)) ? dropObj.amount : 1;
    var effect = dropObj.iterationEffect;
    var result = 0;
    TAA.ed.functions.log(1, "Nominal drop chance: " + chance, "Actual drop chance: " + chance * this.dropItemRate());
    if(!effect || !effect.type || effect.type === 0) {
        if(chance * this.dropItemRate() >= Math.random())
            result = amount;
    }
    else {
        while(amount > 0){
            var test = chance * this.dropItemRate() >= Math.random();
            if(test)
                result++;
            if(effect.type && !isNaN(effect.attenuation) && effect.attenuation !== 0){
                var previousChance = chance;
                chance += effect.attenuation;
                TAA.ed.functions.log(1, "--------------------", "Attenuated chance value: " + previousChance + " ----> " + chance, "--------------------");
            }
            amount--;
        }
    }
    TAA.ed.functions.log(1, "--------------------", "DROP COUNT: " + result, "--------------------");
    return result;
};

/**
 * Auxiliary function to update the returned rewards array
 * 
 * @static
 * @method makeRewardsArray
 */
EnemyDataManager.makeRewardsArray = function(rewards, item, amount){
    if(!item || isNaN(amount) || amount <= 0) return rewards;
    for(var i=0; i<amount; i++){
        rewards.push(item);
    }
    return rewards;
};

/**
 * Creates a normalized drop object rule based on default rules.
 * 
 * @static
 * @method copyDefaultDropRule
 */
EnemyDataManager.copyDefaultDropRule = function(){
    var dropObj = {};
    try{
        dropObj.base = parseInt(TAA.ed.Parameters.Default.drops['Base Chance']);
        dropObj.conditions = JSON.parse(TAA.ed.Parameters.Default.drops['Drop Conditions']);
        if(TAA.ed.Parameters.Default.drops['Iteration Effect']){
            dropObj.iterationEffect = {};
            var iter = JSON.parse(TAA.ed.Parameters.Default.drops['Iteration Effect']);
            dropObj.iterationEffect.type = parseInt(iter['Type']);
            var value = parseInt(iter['Attenuation']);
            if(value < 0) value = Math.max(-100, value)/100;
            else value = Math.min(100, value)/100;
            dropObj.iterationEffect.attenuation = value;
        }
    } catch(e){
        console.error("TAA_EnemyDrops - copyDefaultDropRule: Failure to parse default rules.");
        console.error(TAA.ed.Parameters.Default.drops);
        dropObj = undefined;
    }
    return dropObj;
};

/**
 * Deduplicate engine drop settings, so that if the same item is listed more than once,
 * it becomes a single object for drop analysis, and prepare default drop rules for each
 * deduplicated drop.
 * 
 * @static
 * @method deduplicateDropList
 */
EnemyDataManager.deduplicateDropList = function(dropList){
    if(!dropList || dropList.length <= 0)
        return [];
    var resultList = [];
    var processedItems = [];
    for(var i=0; i<dropList.length; i++){
        var dropObj = EnemyDataManager.copyDefaultDropRule(TAA.ed.Parameters.Default.drops);
        if(!dropObj) break;
        var drop = dropList[i];
        var index = processedItems.indexOf(drop);
        if(index < 0){
            if(DataManager.isItem(drop)){
                dropObj.item = drop.id;
                dropObj.type = "item";
            }
            else if(DataManager.isWeapon(drop)){
                dropObj.weapon = drop.id;
                dropObj.type = 'weapon';
            }
            else if(DataManager.isArmor(drop)){
                dropObj.armor = drop.id;
                dropObj.type = 'armor';
            }
            dropObj.amount = 1;
            resultList.push(dropObj);
            processedItems.push(drop);
        }
        else{
            resultList[index].amount++;
        }
    }
    return resultList;
};

/**
 * Determines rewards by going through default drop rule settings for the given enemy, 
 * if the parameter is set. Otherwise, just return the loot set through the engine.
 * 
 * @static
 * @method makeDefaultDropItems
 */
EnemyDataManager.makeDefaultDropItems = function(enemy){
    if(!TAA.ed.Parameters.Default.drops || (Object.keys(TAA.ed.Parameters.Default.drops) || []).length <= 0)
        return TAA.ed.alias.Game_Enemy.makeDropItems.call(enemy);
    var dropList = enemy.engineDropList();
    if(!dropList || dropList.length <= 0)
        return [];
    var start = new Date();
    TAA.ed.functions.log(2, "START (Default)::: " + start.getMinutes() + ":" + start.getSeconds() + "." + start.getMilliseconds());
    var items = [];
    var resultArray = [];
    var dropArray = this.deduplicateDropList(dropList);
    for(var j=0; j<dropArray.length; j++){
        var dropObj = dropArray[j];
        var amount = dropObj.amount;
        var chance = this.processItemDropRules(enemy, dropObj);
        resultArray[j] = this.applyDropAttenuation(chance, dropObj);
        TAA.ed.functions.log(1, "Default Drops - amount: " + amount, "chance: " + chance, "dropResult: " + resultArray[j], "item: " + dropObj.item + " / armor: " + dropObj.armor + " / weapon: " + dropObj.weapon);
        if(resultArray[j]){
            if(['item', 'all'].contains(dropObj.type.toLowerCase()) && !isNaN(dropObj.item) && $dataItems[dropObj.item])
                items = this.makeRewardsArray(items, $dataItems[dropObj.item], resultArray[j]);
            if(['armor', 'all'].contains(dropObj.type.toLowerCase()) && !isNaN(dropObj.armor) && $dataArmors[dropObj.armor])
                items = this.makeRewardsArray(items, $dataArmors[dropObj.armor], resultArray[j]);
            if(['weapon', 'all'].contains(dropObj.type.toLowerCase()) && !isNaN(dropObj.weapon) && $dataWeapons[dropObj.weapon])
                items = this.makeRewardsArray(items, $dataWeapons[dropObj.weapon], resultArray[j]);
        }
    }

    var end = new Date();
    TAA.ed.functions.log(2, "END (Default)::: " + end.getMinutes() + ":" + end.getSeconds() + "." + end.getMilliseconds());
    TAA.ed.functions.log(1, "TIME ELAPSED (Default): " + (end.getTime() -start.getTime()));
    return items;
};

/**
 * Determines rewards by going through all drop rule settings for the given enemy
 * 
 * @static
 * @method makeDropItems
 */
EnemyDataManager.makeDropItems = function(enemy){
    var start = new Date();
    TAA.ed.functions.log(2, "START::: " + start.getMinutes() + ":" + start.getSeconds() + "." + start.getMilliseconds());
    if(!enemy._extData && !enemy._extData.drops) return [];
    var items = [];
    var resultArray = [];
    for(var j=0; j<enemy._extData.drops.length; j++){
        var dropObj = enemy._extData.drops[j];
        var amount = isNaN(dropObj.amount) ? 1 : dropObj.amount;
        var chance = this.processItemDropRules(enemy, dropObj);
        resultArray[j] = this.applyDropAttenuation(chance, dropObj);
        TAA.ed.functions.log(1, "amount: " + amount, "chance: " + chance, "dropResult: " + resultArray[j], "item: " + dropObj.item + " / armor: " + dropObj.armor + " / weapon: " + dropObj.weapon);
        if(resultArray[j]){
            if(['item', 'all'].contains(dropObj.type.toLowerCase()) && !isNaN(dropObj.item) && $dataItems[dropObj.item])
                items = this.makeRewardsArray(items, $dataItems[dropObj.item], resultArray[j]);
            if(['armor', 'all'].contains(dropObj.type.toLowerCase()) && !isNaN(dropObj.armor) && $dataArmors[dropObj.item])
                items = this.makeRewardsArray(items, $dataArmors[dropObj.armor], resultArray[j]);
            if(['weapon', 'all'].contains(dropObj.type.toLowerCase()) && !isNaN(dropObj.weapon) && $dataWeapons[dropObj.weapon])
                items = this.makeRewardsArray(items, $dataWeapons[dropObj.weapon], resultArray[j]);
        }
    }

    var end = new Date();
    TAA.ed.functions.log(2, "END::: " + end.getMinutes() + ":" + end.getSeconds() + "." + end.getMilliseconds());
    TAA.ed.functions.log(1, "TIME ELAPSED: " + (end.getTime() -start.getTime()));
    return items;
};

/**
 * Calculates drop rate modifiers. It supports the plugin modifier variable and the
 * default engine party ability to double the drop rate (both methods are stackable)
 * 
 * @static
 * @method dropItemRate
 */
EnemyDataManager.dropItemRate = function(){
    var dropRate = $gameParty.hasDropItemDouble() ? 2 : 1;
    if(TAA.ed.Parameters.Default.DropRateVar > 0){
        var rateVar = $gameVariables.value(TAA.ed.Parameters.Default.DropRateVar);
        dropRate += rateVar/100;
    }
    return dropRate;
}

/**
 * Parse enemy notes looking for drop conditions settings
 * 
 * @static
 * @method parseEnemyDropNotes
 */
EnemyDataManager.parseEnemyDropNotes = function(enemyId){
    if($dataEnemies[enemyId] === undefined) return undefined;
    var note = $dataEnemies[enemyId].note;
    if(note === undefined) return undefined;
    var lines = note.split(/[\r\n]+/);
    var drops = [];
    var data = {};
    var isTag = false;
    for(var i=0; i<lines.length; i++){
        if(lines[i].match(/<TAA_ED:\s*(item|armor|weapon)\s*=\s*(\d+|v\[\d+\]|[^\s]+)\s+(?:,\s*amount\s*=\s*\d+|,\s*attenuation=\s*\d+\s*(?:;\s*\d+)?)\s*/i)){
            data = new Object();
            var type = (RegExp.$1).toLowerCase();
            switch(type){
                case 'item':
                    var item = this.getVarValueFromText(RegExp.$2, this._itemIndex);
                    break;
                case 'armor':
                    var item = this.getVarValueFromText(RegExp.$2, this._armorIndex);
                    break;
                case 'weapon':
                    var item = this.getVarValueFromText(RegExp.$2, this._weaponIndex);
                    break;
            }
            var amount = 1;
            if(lines[i].match(/amount\s*=\s*(\d+)/i)){
                amount = parseInt(RegExp.$1);
            }
            var attenuationType = 0;
            var attenuationValue = 0;
            /*if(lines[i].match(/attenuation=\s*(\d+)\s*(?:;\s*(\d+))?)/i)){
                var attenuationType = parseInt(RegExp.$1);
                var attenuationValue = !isNaN(RegExp.$2) ? parseInt(RegExp.$2) : 0;
                if(attenuationType < 0 || attenuationType > 2) attenuationType = 0;
            }*/
            if (lines[i].match(/attenuation=\s*(\d+)\s*(?:;\s*(\d+))?/i)) {
                var attenuationType = parseInt(RegExp.$1);
                var attenuationValue = !isNaN(RegExp.$2) ? parseInt(RegExp.$2) : 0;
                if (attenuationType < 0 || attenuationType > 2) attenuationType = 0;
            }
            ////
            data.type = type;
            data[type] = item;
            data.amount = amount;
            data.iterationEffect = {
                type: attenuationType,
                attenuation: attenuationValue
            };
            data.conditions = [];
            isTag = true;
        }
        else if(lines[i].match(/<\/TAA_ED>/i)){
            isTag = false;
            drops.push(data);
        }
        else if(isTag){
            data.conditions.push(lines[i]);
        }
    }
    return drops;
};

//=============================================================================
// Game_System
//=============================================================================

TAA.ed.alias.Game_System = TAA.ed.alias.Game_System || {};
TAA.ed.alias.Game_System.initialize = Game_System.prototype.initialize;
Game_System.prototype.initialize = function(){
    TAA.ed.alias.Game_System.initialize.call(this);
    this._edChanceProcessing = TAA.ed.Parameters.Default.ChanceProcessing;
    this._edUseEngineDrops = TAA.ed.Parameters.Default.EngineSettings;
};

Game_System.prototype.edChanceProcessing = function(){
    return this._edChanceProcessing;
};

Game_System.prototype.edUpdateChanceProcessing = function(option){
    if([0,1].contains(option))
        this._edChanceProcessing = option;
};

Game_System.prototype.edUseEngineSettings = function(){
    return this._edUseEngineDrops;
};

Game_System.prototype.edEnableEngineSettings = function(){
    this._edUseEngineDrops = true;
};

Game_System.prototype.edDisableEngineSettings = function(){
    this._edUseEngineDrops = false;
};

//=============================================================================
// Game_Interpreter (MV)
//=============================================================================

TAA.ed.alias_Game_Interpreter = TAA.ed.alias_Game_Interpreter || {};
TAA.ed.alias_Game_Interpreter.pluginCommand = Game_Interpreter.prototype.pluginCommand;
Game_Interpreter.prototype.pluginCommand = function(command, args){
    TAA.ed.alias_Game_Interpreter.pluginCommand.call(this, command, args);
    if(command.toLowerCase() === 'enemydrops'){
        if(args[0] && args[0].toLowerCase() === 'chance'){
            if(args[1] && args[1].toLowerCase() === 'death'){
                $gameSystem.edUpdateChanceProcessing(0);
            }
            else if(args[1] && args[1].toLowerCase() === 'battle'){
                $gameSystem.edUpdateChanceProcessing(1);
            }
        }
        else if(args[0] && args[0].toLowerCase() === 'engine'){
            if(args[1] && args[1].toLowerCase() === 'enable'){
                $gameSystem.edEnableEngineSettings();
            }
            else if(args[1] && args[1].toLowerCase() === 'disable'){
                $gameSystem.edDisableEngineSettings();
            }
        }
    }
};

//=============================================================================
// Plugin Commands (MZ)
//=============================================================================

if(Utils.RPGMAKER_NAME === 'MZ'){
    PluginManager.registerCommand(TAA.ed.PluginName, 'chanceProcessing', args => {
        const type = parseInt(args.process);
        $gameSystem.edUpdateChanceProcessing(type);
    });

    PluginManager.registerCommand(TAA.ed.PluginName, 'engineSettings', args => {
        const settings = args.Settings === true;
        if(settings)
            $gameSystem.edEnableEngineSettings();
        else
            $gameSystem.edDisableEngineSettings();
    });
}

//=============================================================================
// BattleManager
//=============================================================================

TAA.ed.alias.BattleManager = TAA.ed.alias.BattleManager || {};
TAA.ed.alias.BattleManager.invokeAction = BattleManager.invokeAction;
BattleManager.invokeAction = function(subject, target){
    TAA.ed.alias.BattleManager.invokeAction.call(this, subject, target);
    if(DataManager.isSkill(this._action.item())){
        if(target.isEnemy()){
            target.edAddSkillHit(this._action.item());
            if(subject.isActor()) target.edAddActorStrikes(subject);
        }
        if(subject.isEnemy()){
            subject.edAddSkillUse(this._action.item());
        }
    }
    if(DataManager.isItem(this._action.item())){
        if(target.isEnemy()){
            target.edAddItemUses(this._action.item());
        }
        if(subject.isActor()){
            $gameParty.edAddItemUses(this._action.item());
        }
    }
};

TAA.ed.alias.BattleManager.makeRewards = BattleManager.makeRewards;
BattleManager.makeRewards = function(){
    TAA.ed.alias.BattleManager.makeRewards.call(this);
    $gameParty.edClearItemUses();
};

TAA.ed.alias.BattleManager.displayDropItems = BattleManager.displayDropItems;
BattleManager.displayDropItems = function(){
    if(!TAA.ed.Parameters.BattleMsg.Summarize){
        TAA.ed.alias.BattleManager.displayDropItems.call(this);
    }
    else{
        var items = {};
        var list = this._rewards.items;
        for(var i=0; i<list.length; i++){
            if(items[list[i].name]){
                items[list[i].name].amount++;
            }
            else{
                items[list[i].name] = {
                    name:list[i].name,
                    amount:1
                };
            }
        }
        var keys = Object.keys(items);
        if(keys.length > 0){
            $gameMessage.newPage();
            keys.forEach(function(i) {
                var text = TAA.ed.Parameters.BattleMsg.Format.format(items[i].name, items[i].amount);
                $gameMessage.add(TextManager.obtainItem.format(text));
            }, items);
        }
    }
};

//=============================================================================
// Game_Enemy
//=============================================================================

TAA.ed.alias.Game_Enemy = TAA.ed.alias.Game_Enemy || {};
TAA.ed.alias.Game_Enemy.initMembers = Game_Enemy.prototype.initMembers;
Game_Enemy.prototype.initMembers = function(){
    TAA.ed.alias.Game_Enemy.initMembers.call(this);
    this._edData = {
        skillsUsedOrder: [],
        skillsUsedCount: [],
        skillsHitOrder: [],
        skillsHitCount: [],
        stypeOrder: [],
        stypeCount: [],
        elementOrder: [],
        elementCount: [],
        actorOrder: [],
        actorCount: [],
        itemOrder: [],
        itemCount: [],
        deathTurn: 0,
        lastDmg: 0,
        maxDmg: 0
    };
    this._extData = this._extData || {};
    this._dropRewards = [];
};

Game_Enemy.prototype.edAddSkillHit = function(skill){
    if(!DataManager.isSkill(skill)) return;
    this._edData.skillsHitOrder.push(skill.id);
    this._edData.stypeOrder.push(skill.stypeId);
    this._edData.skillsHitCount[skill.id] = this._edData.skillsHitCount[skill.id] || 0;
    this._edData.skillsHitCount[skill.id]++;
    this._edData.skillsHitCount[skill.stypeId] = this._edData.skillsHitCount[skill.stypeId] || 0;
    this._edData.skillsHitCount[skill.stypeId]++;
    if(skill.elementId && skill.elementId > 0){
        this._edData.elementOrder.push(skill.elementId);
        this._edData.elementCount[skill.elementId] = this._edData.elementCount[skill.elementId] || 0;
        this._edData.elementCount[skill.elementId]++;
    }
};

Game_Enemy.prototype.edAddSkillUse = function(skill){
    if(!DataManager.isSkill(skill)) return;
    this._edData.skillsUsedOrder.push(skill.id);
    this._edData.skillsUsedCount[skill.id] = this._edData.skillsUsedCount[skill.id] || 0;
    this._edData.skillsUsedCount[skill.id]++;
};

Game_Enemy.prototype.edAddActorStrikes = function(actor){
    if(actor.constructor.name !== 'Game_Actor') return;
    this._edData.actorOrder.push(actor.actorId());
    this._edData.actorCount[actor.actorId()] = this._edData.actorCount[actor.actorId()] || 0;
    this._edData.actorCount[actor.actorId()]++;
};

Game_Enemy.prototype.edAddItemUses = function(item){
    if(!DataManager.isItem(item)) return;
    this._edData.itemOrder.push(item.id);
    this._edData.itemCount[item.id] = this._edData.itemCount[item.id] || 0;
    this._edData.itemCount[item.id]++;
};

Game_Enemy.prototype.edSaveDamage = function(value){
    this._edData.lastDmg = value;
    this._edData.maxDmg = Math.max(this._edData.maxDmg, value);
};

Game_Enemy.prototype.edDeathTurn = function(){
    return Math.max(0, this._edData.deathTurn);
};

Game_Enemy.prototype.edLastDamage = function(){
    return Math.max(0, this._edData.lastDmg);
};

Game_Enemy.prototype.edMaxDamage = function(){
    return Math.max(0, this._edData.maxDmg);
};

Game_Enemy.prototype.edDataGetLastValue = function(array){
    if(!array) return 0;
    return Math.max(0, array[array.length-1]);
};

Game_Enemy.prototype.edDataGetCountValue = function(array, id){
    if(!array || isNaN(id)) return 0;
    if(array[id]){
        return array[id];
    }
    return 0;
};

Game_Enemy.prototype.edLastSkillUsed = function(){
    return this.edDataGetLastValue(this._edData.skillsUsedOrder);
};

Game_Enemy.prototype.edSkillUsedCount = function(id){
    return this.edDataGetCountValue(this._edData.skillsUsedCount, id);
};

Game_Enemy.prototype.edLastSkillHit = function(){
    return this.edDataGetLastValue(this._edData.skillsHitOrder);
};

Game_Enemy.prototype.edSkillHitCount = function(id){
    return this.edDataGetCountValue(this._edData.skillsHitCount, id);
};

Game_Enemy.prototype.edLastSkillTypeHit = function(){
    return this.edDataGetLastValue(this._edData.stypeOrder);
};

Game_Enemy.prototype.edSkillTypeHitCount = function(id){
    return this.edDataGetCountValue(this._edData.stypeCount, id);
};

Game_Enemy.prototype.edLastElementHit = function(){
    return this.edDataGetLastValue(this._edData.elementOrder);
};

Game_Enemy.prototype.edElementHitCount = function(id){
    return this.edDataGetCountValue(this._edData.elementCount, id);
};

Game_Enemy.prototype.edLastActorHit = function(){
    return this.edDataGetLastValue(this._edData.actorOrder);
};

Game_Enemy.prototype.edActorHitCount = function(id){
    return this.edDataGetCountValue(this._edData.actorCount, id);
};

Game_Enemy.prototype.edLastItemHit = function(){
    return this.edDataGetLastValue(this._edData.itemOrder);
};

Game_Enemy.prototype.edItemHitCount = function(id){
    return this.edDataGetCountValue(this._edData.itemCount, id);
};

Game_Enemy.prototype.engineDropList = function(){
    return this.enemy().dropItems.reduce(function(r, di){
        var item = this.itemObject(di.kind, di.dataId);
        if(item !== null)
            return r.concat(item);
        else
            return r;
    }.bind(this), []);
};

TAA.ed.alias.Game_Enemy.makeDropItems = Game_Enemy.prototype.makeDropItems;
Game_Enemy.prototype.makeDropItems = function(){
    if(TAA.ed.Parameters.Default.ChanceProcessing === 1){
        this.prepareDropRewards();
    }
    if(!this._dropRewards) return [];
    return this._dropRewards;
};

Game_Enemy.prototype.prepareDropRewards = function(){
    var list1 = [];
    var list2 = [];
    if($gameSystem.edUseEngineSettings())
        list1 = EnemyDataManager.makeDefaultDropItems(this);
    list2 = EnemyDataManager.makeDropItems(this);

    this._dropRewards = [].concat(list1, list2, this._dropRewards);
};

TAA.ed.alias.Game_Enemy.die = Game_Enemy.prototype.die;
Game_Enemy.prototype.die = function(){
    TAA.ed.alias.Game_Enemy.die.call(this);
    this._edData.deathTurn = $gameTroop.turnCount();
    if(TAA.ed.Parameters.Default.ChanceProcessing === 0) 
        this.prepareDropRewards();
};

TAA.ed.alias.Game_Enemy.setup = Game_Enemy.prototype.setup;
Game_Enemy.prototype.setup = function(enemyId, x, y){
    TAA.ed.alias.Game_Enemy.setup.call(this, enemyId, x, y);
    this._extData = this.setupExtendedDropData(enemyId);
};

Game_Enemy.prototype.setupExtendedDropData = function(enemyId){
    if((Object.keys(this._extData) || []).length <= 0)
        var data = EnemyDataManager.getNormalizedData(enemyId);
    else
        var data = this._extData;
    var customData = EnemyDataManager.parseEnemyDropNotes(enemyId);
    data.drops = (customData || []).length > 0 ? customData : data.drops;
    return data;
};

TAA.ed.alias.Game_Enemy.gainHP = Game_Enemy.prototype.gainHp;
Game_Enemy.prototype.gainHp = function(value){
    if(value < 0){
        this.edSaveDamage(-value);
    }
    TAA.ed.alias.Game_Enemy.gainHP.call(this, value);
};

//=============================================================================
// Game_Party
//=============================================================================

TAA.ed.alias.Game_Party = TAA.ed.alias.Game_Party || {};
TAA.ed.alias.Game_Party.initialize = Game_Party.prototype.initialize;
Game_Party.prototype.initialize = function(){
    TAA.ed.alias.Game_Party.initialize.call(this);
    this._edBattleItems = {
        ordered: [],
        count: []
    };
};

Game_Party.prototype.edAddItemUses = function(item){
    if(!DataManager.isItem(item)) return;
    this._edBattleItems.ordered.push(item.id);
    this._edBattleItems.count[item.id] = this._edBattleItems.count[item.id] || 0;
    this._edBattleItems.count[item.id]++;
};

Game_Party.prototype.edClearItemUses = function(){
    this._edBattleItems = {
        ordered: [],
        count: []
    };
};

Game_Party.prototype.edLastItemUsed = function(){
    if(!this._edBattleItems.ordered) return 0;
    return Math.max(0, this._edBattleItems.ordered[this._edBattleItems.ordered.length-1]);
};

Game_Party.prototype.edItemUsedCount = function(id){
    if(!this._edBattleItems.ordered || isNaN(id)) return 0;
    if(this._edBattleItems.ordered[id]){
        return this._edBattleItems.ordered[id];
    }
    return 0;
};

if(!TAA.el){
    Game_Party.prototype.lowestBattlerLevel = function() {
        return Math.min.apply(null, this.battleMembers().map(function(actor) {
            return actor.level;
        }));
    };
    
    Game_Party.prototype.highestBattlerLevel = function() {
        return Math.max.apply(null, this.battleMembers().map(function(actor) {
            return actor.level;
        }));
    };

    Game_Party.prototype.averageBattlerLevel = function() {
        var sum = 0;
        var party = this.battleMembers();
        var size = party.length;
        for(var i=0; i < size; i++){
            sum += party[i].level;
        }
        return Math.round(sum/size);
    };
    
    Game_Party.prototype.partyActiveStates = function() {
        var list = [];
        var leader = this.leader();
        list = leader.states().slice();
        var followers = $gamePlayer.followers();
        for(var i=0; i<followers.length; i++){
            list = list.concat(followers[i].states());
        }
        return list;
    };
}